#!/usr/bin/env python

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

# Generated by Django 1.9.7 on 2016-06-21 15:11
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('userprofile', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='userprofile',
            name='sql_scratchpad',
            field=models.TextField(default='', verbose_name='SQL scratchpad for query builder'),  # noqa
            preserve_default=False,
        ),
    ]



#!/usr/bin/env python
# crate_anon/nlp_manager/regex_units.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

from typing import List, Optional, Tuple
from crate_anon.nlp_manager.regex_test import test_text_regex
from crate_anon.nlp_manager.regex_numbers import (
    BILLION,
    MULTIPLY_OR_SPACE,
    PLAIN_INTEGER,
    POWER,
)


# =============================================================================
# Physical units
# =============================================================================

OUT_OF_SEPARATOR = r"(?: \/ | \b out \s+ of \b )"


def per(numerator: str, denominator: str,
        include_power_minus1: bool = True) -> str:
    # Copes with blank/optional numerators, too.
    options = [
        r"(?: {numerator} \s* (?: \/ | \b per \b) \s* {denominator} )".format(
            numerator=numerator, denominator=denominator),
    ]
    if include_power_minus1:
        options.append(
            r"(?: {numerator} \s* \b {denominator} \s* -1 )".format(
                numerator=numerator, denominator=denominator))
    return r"(?: {} )".format(r" | ".join(options))
    # Use of "\s* \b" rather than "\s+" is so we can have a BLANK numerator.


def _out_of_str(n_as_regex: str):
    # / n
    # out of n
    return r"(?: {out_of} \s* {n} \b)".format(out_of=OUT_OF_SEPARATOR,
                                              n=n_as_regex)


def out_of(n: int) -> str:
    return _out_of_str(str(n))


def out_of_anything() -> str:
    # out_of(n) where n is any number
    return _out_of_str(PLAIN_INTEGER)


def power(x: str, n: int, allow_no_operator: bool=False) -> str:
    return r"(?: {x} \s* {power}{optional} \s* {n})".format(
        x=x,
        power=POWER,
        optional="?" if allow_no_operator else "",
        n=n,
    )


def units_times(*args: str) -> str:
    """For units, where they are notionally multiplied."""
    multiply = MULTIPLY_OR_SPACE + "?"
    joined = multiply.join(args)
    return r"(?: {} )".format(joined)


def units_by_dimension(*args: Tuple[Tuple[str, int], ...],
                       allow_no_operator: bool=False) -> str:
    multiply = " " + MULTIPLY_OR_SPACE + " "
    power_elements = []  # type: List[str]
    for i, unit_exponent in enumerate(args):
        unit, exponent = unit_exponent
        assert(exponent != 0)
        power_elements.append(
            power(unit, exponent, allow_no_operator=allow_no_operator))
    joined_power_elements = multiply.join(power_elements)
    power_style = r"(?: {} )".format(joined_power_elements)
    options = [power_style]
    # noinspection PyChainedComparisons
    if len(args) == 2 and args[0][1] > 0 and args[1][1] < 0:
        # x per y
        options.append(per(args[0][0], args[1][0],
                           include_power_minus1=False))
    return r"(?: {} )".format(r" | ".join(options))


# -----------------------------------------------------------------------------
# Distance
# -----------------------------------------------------------------------------

M = r"(?: met(?:re|er)s? | m )"  # m, metre(s), meter(s)
CM = r"(?: cm | centimet(?:re|er)s? )"   # cm, centimetre(s), centimeter(s)
MM = r"(?: mm | millimet(?:re|er)s? )"   # mm, millimetre(s), millimeter(s)

FEET = r"""(?: f(?:ee|oo)?t | \' | ’ | ′ )"""
# ... feet, foot, ft
# ... apostrophe, right single quote (U+2019), prime (U+2032)
INCHES = r'''(?: in(?:ch(?:e)?)?s? | \" | ” | ″)'''
# ... in, ins, inch, inches, [inchs = typo but clear]
# ... ", right double quote (U+2014), double prime (U+2033)

# -----------------------------------------------------------------------------
# Mass
# -----------------------------------------------------------------------------

MCG = r"(?: mcg | microgram(?:me)?s? | [μu]g )"  # you won't stop people using ug...  # noqa
MG = r"(?: mg | milligram(?:me)?s? )"  # mg, milligram, milligrams, milligramme, milligrammes  # noqa
G = r"(?: gram(?:me)?s? | g )"  # g, gram, grams, gramme, grammes  # noqa
KG = r"(?: kgs? | kilo(?:gram(?:me)?)?s? )"  # kg, kgs, kilos ... kilogrammes etc.  # noqa
LB = r"(?: pounds? | lbs? )"  # pound(s), lb(s)
STONES = r"(?: stones? | st\.? )"  # stone(s), st, st.

# -----------------------------------------------------------------------------
# Volume
# -----------------------------------------------------------------------------

L = r"(?: lit(?:re|er)s? | L )"  # L, litre(s), liter(s)
DL = r"(?: d(?:eci)?{L} )".format(L=L)
ML = r"(?: m(?:illi)?{L} )".format(L=L)
CUBIC_MM = r"""(?: (?:\b cubic \s+ {mm}) | {mm_cubed} )""".format(
    mm=MM,
    mm_cubed=power(MM, 3, allow_no_operator=True)
)
# cubic mm, etc. | mm^3, mm3, mm 3, etc.

# -----------------------------------------------------------------------------
# Inverse volume
# -----------------------------------------------------------------------------

PER_CUBIC_MM = per("", CUBIC_MM)

# -----------------------------------------------------------------------------
# Time
# -----------------------------------------------------------------------------

HOUR = r"(?:h(?:rs?|ours?)?)"   # h, hr, hrs, hour, hours

# -----------------------------------------------------------------------------
# Counts, proportions
# -----------------------------------------------------------------------------

PERCENT = r"""(?:%|pe?r?\s?ce?n?t)"""
# must have pct, other characters optional

# -----------------------------------------------------------------------------
# Arbitrary count things
# -----------------------------------------------------------------------------

CELLS = r"(?:\b cells? \b)"
OPTIONAL_CELLS = CELLS + "?"
MILLIMOLES = r"(?: m(?:illi)?mol(?:es?)? )"
MILLIEQ = r"(?:m(?:illi)?Eq)"

UNITS = r"(?:[I]?U)"  # U units, IU international units
MILLIUNITS = r"(?:m[I]?U)"
MICROUNITS = r"(?:[μu][I]?U)"

SCORE = r"(?:scored?)"  # score(d)

# -----------------------------------------------------------------------------
# Concentration
# -----------------------------------------------------------------------------

MILLIMOLAR = r"(?:mM)"  # NB case-insensitive... confusable with millimetres
MG_PER_DL = per(MG, DL)
MG_PER_L = per(MG, L)
MILLIMOLES_PER_L = per(MILLIMOLES, L)
MILLIEQ_PER_L = per(MILLIEQ, L)
BILLION_PER_L = per(BILLION, L)
CELLS_PER_CUBIC_MM = per(OPTIONAL_CELLS, CUBIC_MM)

MILLIUNITS_PER_L = per(MILLIUNITS, L)
MICROUNITS_PER_ML = per(MICROUNITS, ML)

# -----------------------------------------------------------------------------
# Speed
# -----------------------------------------------------------------------------

MM_PER_H = per(MM, HOUR)

# -----------------------------------------------------------------------------
# Pressure
# -----------------------------------------------------------------------------

MM_HG = r"(?: mm \s* Hg )"  # mmHg, mm Hg
# ... likelihood of "millimetres of mercury" quite small?

# -----------------------------------------------------------------------------
# Things to powers
# -----------------------------------------------------------------------------

SQ_M = r"""
    (?:  # square metres
        (?: sq(?:uare)? \s+ {m} )       # sq m, square metres, etc.
        | (?: {m} \s+ sq(?:uared?)? )   # m sq, metres square(d), etc.
        | {m_sq}                        # m ^ 2, etc.
    )
""".format(m=M, m_sq=power(M, 2))

# BMI
KG_PER_SQ_M = r"(?: {kg_per_sqm} | {kg_sqm_pow_minus2} )".format(
    kg_per_sqm=per(KG, SQ_M, include_power_minus1=False),
    kg_sqm_pow_minus2=units_times(KG, power(M, -2)),
)


# =============================================================================
#  Generic conversion functions
# =============================================================================

def kg_from_st_lb_oz(stones: float = 0,
                     pounds: float = 0,
                     ounces: float = 0) -> Optional[float]:
    # 16 ounces in a pound
    # 14 pounds in a stone
    # 1 avoirdupois pound = 0.45359237 kg
    # https://en.wikipedia.org/wiki/Pound_(mass)
    # Have you the peas? "Goods of weight"; aveir de peis (OFr.; see OED).
    try:
        total_pounds = (stones * 14) + pounds + (ounces / 16)
        return 0.45359237 * total_pounds
    except (TypeError, ValueError):
        return None


def m_from_ft_in(feet: float = 0, inches: float = 0) -> Optional[float]:
    # 12 inches in a foot
    # 1 inch = 25.4 mm
    try:
        total_inches = (feet * 12) + inches
        return total_inches * 25.4 / 1000
    except (TypeError, ValueError):
        return None


def m_from_m_cm(metres: float = 0, centimetres: float = 0) -> Optional[float]:
    try:
        return metres + (centimetres / 100)
    except (TypeError, ValueError):
        return None


def assemble_units(components: List[Optional[str]]) -> str:
    """Takes e.g. ["ft", "in"] and makes "ft in"."""
    active_components = [c for c in components if c]
    return " ".join(active_components)


# =============================================================================
#  Tests
# =============================================================================

def test_unit_regexes(verbose: bool = False) -> None:
    test_text_regex("per(n, d)", per("n", "d"), [
        ("blah n per d blah", ["n per d"]),
        ("blah n/d blah", ["n/d"]),
        ("n / d", ["n / d"]),
        ("n d -1", ["n d -1"]),
        ("n d -1", ["n d -1"]),
        ("n blah d", []),
    ], verbose=verbose)
    test_text_regex("out_of(5)", out_of(5), [
        ("4 out of 5", ["out of 5"]),
        ("4/5", ["/5"]),
        ("4 / 5", ["/ 5"]),
    ], verbose=verbose)

    test_text_regex("M", M, [
        ("5 metres long", ["metres"]),
        ("5 meters long", ["meters"]),
        ("5m long", ["m"]),
    ], verbose=verbose)
    test_text_regex("CM", CM, [
        ("5 centimetres long", ["centimetres"]),
        ("5 centimeters long", ["centimeters"]),
        ("5cm long", ["cm"]),
    ], verbose=verbose)
    test_text_regex("MM", MM, [
        ("5 millimetres long", ["millimetres"]),
        ("5 millimeters long", ["millimeters"]),
        ("5mm long", ["mm"]),
    ], verbose=verbose)
    test_text_regex("FEET", FEET, [
        ("5 feet long", ["feet"]),
        ("5 foot long", ["foot"]),
        ("5' long", ["'"]),  # ASCII apostrophe
        ("5’ long", ["’"]),  # right single quote (U+2019)
        ("5′ long", ["′"]),  # prime (U+2032)
    ], verbose=verbose)
    test_text_regex("INCHES", INCHES, [
        ("5 inches long", ["inches"]),
        ("5 in long", ["in"]),
        ('5" long', ['"']),  # ASCII double quote
        ("5” long", ["”"]),  # right double quote (U+2014)
        ("5″ long", ["″"]),  # double prime (U+2033)
    ], verbose=verbose)

    test_text_regex("MCG", MCG, [
        ("5 micrograms", ["micrograms"]),
        ("5 mcg", ["mcg"]),
        ("5 ug", ["ug"]),
        ("5 μg", ["μg"]),
    ], verbose=verbose)
    test_text_regex("MG", MG, [
        ("5 milligrams", ["milligrams"]),
        ("5 mg", ["mg"]),
    ], verbose=verbose)
    test_text_regex("G", G, [
        ("5 grams", ["grams"]),
        ("5 g", ["g"]),
    ], verbose=verbose)
    test_text_regex("KG", KG, [
        ("5 kilograms", ["kilograms"]),
        ("5 kg", ["kg"]),
    ], verbose=verbose)
    test_text_regex("LB", LB, [
        ("5 pounds", ["pounds"]),
        ("5 lb", ["lb"]),
    ], verbose=verbose)
    test_text_regex("STONES", STONES, [
        ("5 stones", ["stones"]),
        ("5 stone", ["stone"]),
        ("5 st", ["st"]),
    ], verbose=verbose)

    test_text_regex("L", L, [
        ("5 litres", ["litres"]),
        ("5 liters", ["liters"]),
        ("5 l", ["l"]),
        ("5 L", ["L"]),
    ], verbose=verbose)
    test_text_regex("DL", DL, [
        ("5 decilitres", ["decilitres"]),
        ("5 deciliters", ["deciliters"]),
        ("5 dl", ["dl"]),
        ("5 dL", ["dL"]),
    ], verbose=verbose)
    test_text_regex("ML", ML, [
        ("5 millilitres", ["millilitres"]),
        ("5 milliliters", ["milliliters"]),
        ("5 ml", ["ml"]),
        ("5 mL", ["mL"]),
    ], verbose=verbose)
    test_text_regex("CUBIC_MM", CUBIC_MM, [
        ("5 mm^3", ["mm^3"]),
        ("5 cubic mm", ["cubic mm"]),
        ("5 cubic millimetres", ["cubic millimetres"]),
    ], verbose=verbose)

    test_text_regex("HOUR", HOUR, [
        ("5 hours", ["hours"]),
        ("5 hr", ["hr"]),
        ("5 h", ["h"]),
    ], verbose=verbose)

    test_text_regex("PERCENT", PERCENT, [
        ("5 percent", ["percent"]),
        ("5 per cent", ["per cent"]),
        ("5 pct", ["pct"]),
        ("5%", ["%"]),
    ], verbose=verbose)

    test_text_regex("CELLS", CELLS, [
        ("5 cells", ["cells"]),
        ("5 cell", ["cell"]),
    ], verbose=verbose)

    test_text_regex("MILLIMOLES", MILLIMOLES, [
        ("5 millimoles", ["millimoles"]),
        ("5 millimol", ["millimol"]),
        ("5 mmol", ["mmol"]),
    ], verbose=verbose)
    test_text_regex("MILLIEQ", MILLIEQ, [
        ("5 mEq", ["mEq"]),
    ], verbose=verbose)

    test_text_regex("UNITS", UNITS, [
        ("5 U", ["U"]),
        ("5 IU", ["IU"]),
    ], verbose=verbose)
    test_text_regex("MILLIUNITS", MILLIUNITS, [
        ("5 mU", ["mU"]),
        ("5 mIU", ["mIU"]),
    ], verbose=verbose)
    test_text_regex("MICROUNITS", MICROUNITS, [
        ("5 uU", ["uU"]),
        ("5 μU", ["μU"]),
        ("5 uIU", ["uIU"]),
        ("5 μIU", ["μIU"]),
    ], verbose=verbose)

    test_text_regex("SCORE", SCORE, [
        ("I scored 5", ["scored"]),
        ("MMSE score 5", ["score"]),
    ], verbose=verbose)

    test_text_regex("MILLIMOLAR", MILLIMOLAR, [
        ("5 mM", ["mM"]),
    ], verbose=verbose)

    test_text_regex("MM_HG", MM_HG, [
        ("5 mmHg", ["mmHg"]),
        ("5 mm Hg", ["mm Hg"]),
    ], verbose=verbose)

    test_text_regex("SQ_M", SQ_M, [
        ("5 square metres", ["square metres"]),
        ("5 sq m", ["sq m"]),
        ("5 m^2", ["m^2"]),
    ], verbose=verbose)

    test_text_regex("KG_PER_SQ_M", KG_PER_SQ_M, [
        ("5 kg per square metre", ["kg per square metre"]),
        ("5 kg/sq m", ["kg/sq m"]),
        ("5 kg/m^2", ["kg/m^2"]),
        ("5 kg*m^-2", ["kg*m^-2"]),
    ], verbose=verbose)


def test_all(verbose: bool = False) -> None:
    test_unit_regexes(verbose=verbose)


if __name__ == '__main__':
    test_all()



#!/usr/bin/env python
# crate_anon/anonymise/dbholder.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import logging
from typing import List

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.schema import MetaData

# from crate_anon.common.sqla import get_table_names

log = logging.getLogger(__name__)


# =============================================================================
# Convenience object
# =============================================================================

DB_SAFE_CONFIG_FWD_REF = "DatabaseSafeConfig"


class DatabaseHolder(object):
    def __init__(self,
                 name: str,
                 url: str,
                 srccfg: DB_SAFE_CONFIG_FWD_REF = None,
                 with_session: bool = False,
                 with_conn: bool = True,
                 reflect: bool = True,
                 encoding: str = 'utf-8') -> None:
        self.name = name
        self.srccfg = srccfg
        self.engine = create_engine(url, encoding=encoding)
        self.conn = None
        self.session = None
        self._reflect_on_request = reflect
        self._reflected = False
        self._table_names = []  # type: List[str]
        self._metadata = MetaData(bind=self.engine)
        log.debug(self.engine)  # obscures password

        if with_conn:  # for raw connections
            self.conn = self.engine.connect()
        if with_session:  # for ORM
            self.session = sessionmaker(bind=self.engine)()  # for ORM

    def _reflect(self):
        # Reflection is expensive, so we defer unless required
        if not self._reflect_on_request:
            return
        log.info("Reflecting database: {}".format(self.name))
        # self.table_names = get_table_names(self.engine)
        self._metadata.reflect(views=True)  # include views
        self._table_names = [t.name for t in self._metadata.sorted_tables]
        self._reflected = True

    @property
    def metadata(self):
        if not self._reflected:
            self._reflect()
        return self._metadata

    @property
    def table_names(self):
        if not self._reflected:
            self._reflect()
        return self._table_names



# -*- coding: utf-8 -*-
# Generated by Django 1.9.8 on 2017-02-06 16:17
from __future__ import unicode_literals

import crate_anon.crateweb.consent.models
import crate_anon.crateweb.consent.storage
import crate_anon.crateweb.extra.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('consent', '0005_auto_20170203_1348'),
    ]

    operations = [
        migrations.AlterField(
            model_name='emailattachment',
            name='file',
            field=models.FileField(storage=crate_anon.crateweb.consent.storage.CustomFileSystemStorage(base_url='download_privatestorage', location='/home/rudolf/Documents/code/crate/working/crateweb/crate_filestorage'), upload_to=''),  # noqa
        ),
        migrations.AlterField(
            model_name='leaflet',
            name='pdf',
            field=crate_anon.crateweb.extra.fields.ContentTypeRestrictedFileField(blank=True, storage=crate_anon.crateweb.consent.storage.CustomFileSystemStorage(base_url='download_privatestorage', location='/home/rudolf/Documents/code/crate/working/crateweb/crate_filestorage'), upload_to=crate_anon.crateweb.consent.models.leaflet_upload_to),  # noqa
        ),
        migrations.AlterField(
            model_name='letter',
            name='pdf',
            field=models.FileField(storage=crate_anon.crateweb.consent.storage.CustomFileSystemStorage(base_url='download_privatestorage', location='/home/rudolf/Documents/code/crate/working/crateweb/crate_filestorage'), upload_to=''),  # noqa
        ),
        migrations.AlterField(
            model_name='study',
            name='study_details_pdf',
            field=crate_anon.crateweb.extra.fields.ContentTypeRestrictedFileField(blank=True, storage=crate_anon.crateweb.consent.storage.CustomFileSystemStorage(base_url='download_privatestorage', location='/home/rudolf/Documents/code/crate/working/crateweb/crate_filestorage'), upload_to=crate_anon.crateweb.consent.models.study_details_upload_to),  # noqa
        ),
        migrations.AlterField(
            model_name='study',
            name='subject_form_template_pdf',
            field=crate_anon.crateweb.extra.fields.ContentTypeRestrictedFileField(blank=True, storage=crate_anon.crateweb.consent.storage.CustomFileSystemStorage(base_url='download_privatestorage', location='/home/rudolf/Documents/code/crate/working/crateweb/crate_filestorage'), upload_to=crate_anon.crateweb.consent.models.study_form_upload_to),  # noqa
        ),
    ]



#!/usr/bin/env python
# crate_anon/crateweb/consent/views.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import logging
import mimetypes
from typing import Optional

from django.conf import settings
from django.contrib.auth.decorators import user_passes_test
from django.db import transaction
from django.http import HttpResponse, Http404, HttpResponseForbidden
from django.http.response import HttpResponseBase
from django.http.request import HttpRequest
from django.shortcuts import get_object_or_404, render

from crate_anon.common.contenttypes import CONTENTTYPE_PDF
from crate_anon.common.nhs import generate_random_nhs_number
from crate_anon.crateweb.consent.forms import (
    ClinicianResponseForm,
    SingleNhsNumberForm,
    SuperuserSubmitContactRequestForm,
    ResearcherSubmitContactRequestForm,
)
from crate_anon.crateweb.consent.models import (
    CharityPaymentRecord,
    ClinicianResponse,
    ConsentMode,
    ContactRequest,
    Email,
    EmailAttachment,
    Leaflet,
    Letter,
    lookup_patient,
    make_dummy_objects,
    PatientLookup,
    Study,
    TEST_ID_STR,
)
from crate_anon.crateweb.consent.storage import privatestorage
from crate_anon.crateweb.consent.tasks import (
    finalize_clinician_response,
    test_email_rdbm_task,
)
from crate_anon.crateweb.consent.utils import days_to_years
from crate_anon.crateweb.core.utils import (
    is_developer,
    is_superuser,
)
from crate_anon.crateweb.extra.pdf import (
    serve_concatenated_pdf_from_disk,
    serve_html_or_pdf,
)
from crate_anon.crateweb.extra.serve import serve_buffer, serve_file

log = logging.getLogger(__name__)


# =============================================================================
# Additional validators
# =============================================================================

def validate_email_request(user: settings.AUTH_USER_MODEL,
                           email: Email) -> Optional[HttpResponseForbidden]:
    if user.profile.is_developer:
        return  # the developer sees all
    if email.to_researcher:
        # e-mails to clinicians/patients may be restricted
        if user.is_superuser:
            return  # RDBM can see any e-mail to a researcher
        studies = Study.filter_studies_for_researcher(Study.objects.all(),
                                                      user)
        if email.study in studies:
            return  # this e-mail belongs to this researcher
    elif email.to_patient:
        if user.is_superuser:
            return  # RDBM can see any e-mail to a patient
    return HttpResponseForbidden("Not authorized")


def validate_letter_request(user: settings.AUTH_USER_MODEL,
                            letter: Letter) -> Optional[HttpResponseForbidden]:
    if user.profile.is_developer:
        return  # the developer sees all
    if user.is_superuser:
        return  # RDBM can see any letters generated
    if letter.to_researcher:
        studies = Study.filter_studies_for_researcher(Study.objects.all(),
                                                      user)
        if letter.study in studies:
            return  # this e-mail belongs to this researcher
    return HttpResponseForbidden("Not authorized")


# =============================================================================
# Fetchers
# =============================================================================

def get_contact_request(request: HttpRequest,
                        contact_request_id: str) -> ContactRequest:
    if contact_request_id == TEST_ID_STR:
        return make_dummy_objects(request).contact_request
    return get_object_or_404(
        ContactRequest, id=contact_request_id)  # type: ContactRequest


def get_patient_lookup(request: HttpRequest,
                       patient_lookup_id: str) -> PatientLookup:
    if patient_lookup_id == TEST_ID_STR:
        return make_dummy_objects(request).patient_lookup
    return get_object_or_404(
        PatientLookup, id=patient_lookup_id)  # type: PatientLookup


def get_consent_mode(request: HttpRequest,
                     consent_mode_id: str) -> ConsentMode:
    if consent_mode_id == TEST_ID_STR:
        return make_dummy_objects(request).consent_mode
    return get_object_or_404(
        ConsentMode, id=consent_mode_id)  # type: ConsentMode


# =============================================================================
# Views
# =============================================================================

# noinspection PyUnusedLocal
def study_details(request: HttpRequest, study_id: str) -> HttpResponseBase:
    study = get_object_or_404(Study, pk=study_id)  # type: Study
    if not study.study_details_pdf:
        raise Http404("No details")
    return serve_file(study.study_details_pdf.path,
                      content_type=CONTENTTYPE_PDF,
                      as_inline=True)
study_details.login_required = False


# noinspection PyUnusedLocal
def study_form(request: HttpRequest, study_id: str) -> HttpResponseBase:
    study = get_object_or_404(Study, pk=study_id)  # type: Study
    if not study.subject_form_template_pdf:
        raise Http404("No study form for clinicians to complete")
    return serve_file(study.subject_form_template_pdf.path,
                      content_type=CONTENTTYPE_PDF,
                      as_inline=True)
study_form.login_required = False


# noinspection PyUnusedLocal
def study_pack(request: HttpRequest, study_id: str) -> HttpResponseBase:
    study = get_object_or_404(Study, pk=study_id)  # type: Study
    filenames = filter(None, [
        study.study_details_pdf.path
        if study.study_details_pdf else None,
        study.subject_form_template_pdf.path
        if study.subject_form_template_pdf else None,
    ])
    if not filenames:
        raise Http404("No leaflets")
    return serve_concatenated_pdf_from_disk(
        filenames,
        offered_filename="study_{}_pack.pdf".format(study_id)
    )
study_pack.login_required = False


# noinspection PyUnusedLocal
@user_passes_test(is_superuser)
def download_privatestorage(request: HttpRequest,
                            filename: str) -> HttpResponseBase:
    """Superuser access function, used for admin interface only."""
    fullpath = privatestorage.path(filename)
    content_type = mimetypes.guess_type(filename, strict=False)[0]
    # ... guess_type returns a (content_type, encoding) tuple
    return serve_file(fullpath, content_type=content_type, as_inline=True)


@user_passes_test(is_developer)
def generate_fake_nhs(request: HttpRequest, n: str = 10) -> HttpResponse:
    nhs_numbers = [generate_random_nhs_number()
                   for _ in range(int(n))]
    return render(request, 'generate_fake_nhs.html', {
        'nhs_numbers': nhs_numbers
    })


def view_email_html(request: HttpRequest, email_id: str) -> HttpResponse:
    email = get_object_or_404(Email, pk=email_id)  # type: Email
    validate_email_request(request.user, email)
    return HttpResponse(email.msg_html)


def view_email_attachment(request: HttpRequest,
                          attachment_id: str) -> HttpResponseBase:
    attachment = get_object_or_404(EmailAttachment, pk=attachment_id)  # type: EmailAttachment  # noqa
    validate_email_request(request.user, attachment.email)
    if not attachment.file:
        raise Http404("Attachment missing")
    return serve_file(attachment.file.path,
                      content_type=attachment.content_type,
                      as_inline=True)


@user_passes_test(is_developer)
def test_patient_lookup(request: HttpRequest) -> HttpResponse:
    form = SingleNhsNumberForm(
        request.POST if request.method == 'POST' else None)
    if form.is_valid():
        lookup = lookup_patient(nhs_number=form.cleaned_data['nhs_number'],
                                save=False)
        # Don't use a Form. https://code.djangoproject.com/ticket/17031
        return render(request, 'patient_lookup_result.html',
                      {'lookup': lookup})
    return render(request, 'patient_lookup_get_nhs.html', {'form': form})


# noinspection PyUnusedLocal
def view_leaflet(request: HttpRequest, leaflet_name: str) -> HttpResponseBase:
    leaflet = get_object_or_404(Leaflet, name=leaflet_name)  # type: Leaflet
    if not leaflet.pdf:
        raise Http404("Missing leaflet")
    return serve_file(leaflet.pdf.path,
                      content_type=CONTENTTYPE_PDF,
                      as_inline=True)
view_leaflet.login_required = False


def view_letter(request: HttpRequest, letter_id: str) -> HttpResponseBase:
    letter = get_object_or_404(Letter, pk=letter_id)  # type: Letter
    validate_letter_request(request.user, letter)
    if not letter.pdf:
        raise Http404("Missing letter")
    return serve_file(letter.pdf.path,
                      content_type=CONTENTTYPE_PDF,
                      as_inline=True)


def submit_contact_request(request: HttpRequest) -> HttpResponse:
    if request.user.is_superuser:
        form = SuperuserSubmitContactRequestForm(
            request.POST if request.method == 'POST' else None)
    else:
        form = ResearcherSubmitContactRequestForm(
            request.user,
            request.POST if request.method == 'POST' else None)
    if not form.is_valid():
        return render(request, 'contact_request_submit.html', {'form': form})

    study = form.cleaned_data['study']
    request_direct_approach = form.cleaned_data['request_direct_approach']
    contact_requests = []

    # NHS numbers
    if request.user.is_superuser:
        for nhs_number in form.cleaned_data['nhs_numbers']:
            contact_requests.append(
                ContactRequest.create(
                    request=request,
                    study=study,
                    request_direct_approach=request_direct_approach,
                    lookup_nhs_number=nhs_number))
    # RIDs
    for rid in form.cleaned_data['rids']:
        contact_requests.append(
            ContactRequest.create(
                request=request,
                study=study,
                request_direct_approach=request_direct_approach,
                lookup_rid=rid))
    # MRIDs
    for mrid in form.cleaned_data['mrids']:
        contact_requests.append(
            ContactRequest.create(
                request=request,
                study=study,
                request_direct_approach=request_direct_approach,
                lookup_mrid=mrid))

    # Show results.
    # Don't use a Form. https://code.djangoproject.com/ticket/17031
    return render(request, 'contact_request_result.html', {
        'contact_requests': contact_requests,
    })


def finalize_clinician_response_in_background(
        request: HttpRequest,
        clinician_response: ClinicianResponse) -> HttpResponse:
    clinician_response.finalize_a()  # first part of processing
    transaction.on_commit(
        lambda: finalize_clinician_response.delay(clinician_response.id)
    )  # Asynchronous
    return render(request, 'clinician_confirm_response.html', {
        'clinician_response': clinician_response,
    })


def clinician_response_view(request: HttpRequest,
                            clinician_response_id: str) -> HttpResponse:
    """
    REC DOCUMENTS 09, 11, 13 (B): Web form for clinicians to respond with
    """
    clinician_response = get_object_or_404(
        ClinicianResponse, pk=clinician_response_id)  # type: ClinicianResponse
    contact_request = clinician_response.contact_request
    study = contact_request.study
    patient_lookup = contact_request.patient_lookup
    consent_mode = contact_request.consent_mode

    # Build form.
    # - We have an existing clinician_response and wish to modify it
    #   (potentially).
    # - If the clinician is responding to an e-mail, they will be passing
    #   a couple of parameters (including the token) via GET query parameters.
    #   If they're clicking "Submit", they'll be using POST.
    if request.method == 'GET':
        from_email = True
        clinician_response.response_route = ClinicianResponse.ROUTE_EMAIL
        data = request.GET
    else:
        from_email = False
        clinician_response.response_route = ClinicianResponse.ROUTE_WEB
        data = request.POST
    form = ClinicianResponseForm(instance=clinician_response, data=data)
    # log.debug("Form data: {}".format(form.data))

    # Token valid? Check raw data. Say goodbye otherwise.
    # - The raw data in the form is not influenced by the form's instance.
    if form.data['token'] != clinician_response.token:
        return HttpResponseForbidden(
            "Not authorized. The token you passed doesn't match the one you "
            "were sent.")

    # Already responded?
    if clinician_response.responded:
        passed_to_pt = (clinician_response.response ==
                        ClinicianResponse.RESPONSE_A)
        return render(request, 'clinician_already_responded.html', {
            'clinician_response': clinician_response,
            'consent_mode': consent_mode,
            'contact_request': contact_request,
            'Leaflet': Leaflet,
            'passed_to_pt': passed_to_pt,
            'patient_lookup': patient_lookup,
            'settings': settings,
            'study': study,
        })

    # Is the clinician saying yes or no (direct from e-mail)?
    if (from_email and form.data['email_choice'] in (
            ClinicianResponse.EMAIL_CHOICE_Y,
            ClinicianResponse.EMAIL_CHOICE_N)):
        # We can't use form.save() as the data may not validate.
        # It won't validate because the response/clinician name is blank.
        # We can't write to the form directly. So...
        clinician_response.email_choice = form.data['email_choice']
        if clinician_response.email_choice == ClinicianResponse.EMAIL_CHOICE_Y:
            # Ask RDBM to do the work
            clinician_response.response = ClinicianResponse.RESPONSE_R
        else:
            # Veto on clinical grounds
            clinician_response.response = ClinicianResponse.RESPONSE_B
        return finalize_clinician_response_in_background(request,
                                                         clinician_response)

    # Has the clinician made a decision via the web form?
    if form.is_valid():
        clinician_response = form.save(commit=False)  # return unsaved instance
        return finalize_clinician_response_in_background(request,
                                                         clinician_response)

    # If we get here, we need to offer the form up for editing,
    # and mark it as a web response.
    clinician_involvement_requested = (
        contact_request.clinician_involvement ==
        ContactRequest.CLINICIAN_INVOLVEMENT_REQUESTED)
    clinician_involvement_required_yellow = (
        contact_request.clinician_involvement ==
        ContactRequest.CLINICIAN_INVOLVEMENT_REQUIRED_YELLOW)
    clinician_involvement_required_unknown = (
        contact_request.clinician_involvement ==
        ContactRequest.CLINICIAN_INVOLVEMENT_REQUIRED_UNKNOWN)
    extra_form = contact_request.is_extra_form()
    return render(request, 'clinician_response.html', {
        'clinician_response': clinician_response,
        'ClinicianResponse': ClinicianResponse,
        'consent_mode': consent_mode,
        'contact_request': contact_request,
        'Leaflet': Leaflet,
        'patient_lookup': patient_lookup,
        'settings': settings,
        'study': study,

        'form': form,

        'clinician_involvement_requested': clinician_involvement_requested,
        'clinician_involvement_required_yellow':
        clinician_involvement_required_yellow,
        'clinician_involvement_required_unknown':
        clinician_involvement_required_unknown,
        'option_c_available': clinician_involvement_requested,
        'option_r_available': not extra_form,
        'extra_form': extra_form,
        'unknown_consent_mode': contact_request.is_consent_mode_unknown(),

        'permitted_to_contact_discharged_patients_for_n_days':
            settings.PERMITTED_TO_CONTACT_DISCHARGED_PATIENTS_FOR_N_DAYS,
        'permitted_to_contact_discharged_patients_for_n_years':
            days_to_years(
                settings.PERMITTED_TO_CONTACT_DISCHARGED_PATIENTS_FOR_N_DAYS),
    })

clinician_response_view.login_required = False


# noinspection PyUnusedLocal
def clinician_pack(request: HttpRequest,
                   clinician_response_id: str,
                   token: str) -> HttpResponse:
    clinician_response = get_object_or_404(
        ClinicianResponse, pk=clinician_response_id)  # type: ClinicianResponse
    contact_request = clinician_response.contact_request
    # Check token authentication
    if token != clinician_response.token:
        return HttpResponseForbidden(
            "Not authorized. The token you passed doesn't match the one you "
            "were sent.")
    # Build and serve
    pdf = contact_request.get_clinician_pack_pdf()
    offered_filename = "clinician_pack_{}.pdf".format(clinician_response_id)
    return serve_buffer(pdf,
                        offered_filename=offered_filename,
                        content_type=CONTENTTYPE_PDF,
                        as_attachment=False,
                        as_inline=True)

clinician_pack.login_required = False


# -----------------------------------------------------------------------------
# Draft e-mails
# -----------------------------------------------------------------------------

@user_passes_test(is_developer)
def draft_clinician_email(request: HttpRequest,
                          contact_request_id: str) -> HttpResponse:
    contact_request = get_contact_request(request, contact_request_id)
    return HttpResponse(
        contact_request.get_clinician_email_html(save=False)
    )


@user_passes_test(is_developer)
def draft_approval_email(request: HttpRequest,
                         contact_request_id: str) -> HttpResponse:
    contact_request = get_contact_request(request, contact_request_id)
    return HttpResponse(contact_request.get_approval_email_html())


@user_passes_test(is_developer)
def draft_withdrawal_email(request: HttpRequest,
                           contact_request_id: str) -> HttpResponse:
    contact_request = get_contact_request(request, contact_request_id)
    return HttpResponse(contact_request.get_withdrawal_email_html())


# -----------------------------------------------------------------------------
# Draft letters
# -----------------------------------------------------------------------------

@user_passes_test(is_developer)
def draft_approval_letter(request: HttpRequest,
                          contact_request_id: str,
                          viewtype: str) -> HttpResponse:
    contact_request = get_contact_request(request, contact_request_id)
    html = contact_request.get_approval_letter_html()
    return serve_html_or_pdf(html, viewtype)


@user_passes_test(is_developer)
def draft_withdrawal_letter(request: HttpRequest,
                            contact_request_id: str,
                            viewtype: str) -> HttpResponse:
    contact_request = get_contact_request(request, contact_request_id)
    html = contact_request.get_withdrawal_letter_html()
    return serve_html_or_pdf(html, viewtype)


@user_passes_test(is_developer)
def draft_first_traffic_light_letter(request: HttpRequest,
                                     patient_lookup_id: str,
                                     viewtype: str) -> HttpResponse:
    patient_lookup = get_patient_lookup(request, patient_lookup_id)
    html = patient_lookup.get_first_traffic_light_letter_html()
    return serve_html_or_pdf(html, viewtype)


@user_passes_test(is_developer)
def draft_confirm_traffic_light_letter(request: HttpRequest,
                                       consent_mode_id: str,
                                       viewtype: str) -> HttpResponse:
    consent_mode = get_consent_mode(request, consent_mode_id)
    html = consent_mode.get_confirm_traffic_to_patient_letter_html()
    return serve_html_or_pdf(html, viewtype)


@user_passes_test(is_developer)
def draft_traffic_light_decision_form(request: HttpRequest,
                                      patient_lookup_id: str,
                                      viewtype: str) -> HttpResponse:
    patient_lookup = get_patient_lookup(request, patient_lookup_id)
    html = patient_lookup.get_traffic_light_decision_form()
    return serve_html_or_pdf(html, viewtype)


@user_passes_test(is_developer)
def draft_letter_clinician_to_pt_re_study(request: HttpRequest,
                                          contact_request_id: str,
                                          viewtype: str) -> HttpResponse:
    contact_request = get_contact_request(request, contact_request_id)
    html = contact_request.get_letter_clinician_to_pt_re_study()
    return serve_html_or_pdf(html, viewtype)


@user_passes_test(is_developer)
def decision_form_to_pt_re_study(request: HttpRequest,
                                 contact_request_id: str,
                                 viewtype: str) -> HttpResponse:
    contact_request = get_contact_request(request, contact_request_id)
    html = contact_request.get_decision_form_to_pt_re_study()
    return serve_html_or_pdf(html, viewtype)


@user_passes_test(is_superuser)
def charity_report(request: HttpRequest) -> HttpResponse:
    responses = ClinicianResponse.objects.filter(charity_amount_due__gt=0)
    payments = CharityPaymentRecord.objects.all()
    total_due = sum([x.charity_amount_due for x in responses])
    total_paid = sum([x.amount for x in payments])
    outstanding = total_due - total_paid
    return render(request, 'charity_report.html', {
        'responses': responses,
        'payments': payments,
        'total_due': total_due,
        'total_paid': total_paid,
        'outstanding': outstanding,
    })


@user_passes_test(is_superuser)
def exclusion_report(request: HttpRequest) -> HttpResponse:
    consent_modes = ConsentMode.objects.filter(current=True,
                                               exclude_entirely=True)
    return render(request, 'exclusion_report.html', {
        'consent_modes': consent_modes,
    })


@user_passes_test(is_superuser)
def test_email_rdbm(request: HttpRequest) -> HttpResponse:
    test_email_rdbm_task.delay()
    return render(request, 'test_email_rdbm_ack.html', {
        'settings': settings,
    })



#!/usr/bin/env python
# crate_anon/preprocess/rio_constants.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

# Tables in RiO v6.2 Core:
RIO_TABLE_MASTER_PATIENT = "ClientIndex"
RIO_TABLE_ADDRESS = "ClientAddress"
RIO_TABLE_PROGRESS_NOTES = "PrgProgressNote"
RIO_TABLE_CLINICAL_DOCUMENTS = "ClientDocument"
# Columns in RiO Core:
RIO_COL_PATIENT_ID = "ClientID"  # RiO 6.2: VARCHAR(15)
RIO_COL_NHS_NUMBER = "NNN"  # RiO 6.2: CHAR(10) ("National NHS Number")
RIO_COL_POSTCODE = "PostCode"  # ClientAddress.PostCode
RIO_COL_DEFAULT_PK = "SequenceID"  # INT
RIO_COL_USER_ASSESS_DEFAULT_PK = "type12_NoteID"

# Tables in RiO CRIS Extract Program (RCEP) output database:
RCEP_TABLE_MASTER_PATIENT = "Client_Demographic_Details"
RCEP_TABLE_ADDRESS = "Client_Address_History"
RCEP_TABLE_PROGRESS_NOTES = "Progress_Notes"
# Columns in RCEP extract:
RCEP_COL_PATIENT_ID = "Client_ID"  # RCEP: VARCHAR(15)
RCEP_COL_NHS_NUMBER = "NHS_Number"  # RCEP: CHAR(10)
RCEP_COL_POSTCODE = "Post_Code"  # RCEP: NVARCHAR(10)
# ... general format (empirically): "XX12 3YY" or "XX1 3YY"; "ZZ99" for unknown
# This matches the ONPD "pdcs" format.
RCEP_COL_MANGLED_KEY = "Document_ID"

# CPFT hacks (RiO tables added to RCEP output):
CPFT_RCEP_TABLE_FULL_PROGRESS_NOTES = "Progress_Notes_II"

# Columns in ONS Postcode Database (from CRATE import):
ONSPD_TABLE_POSTCODE = "postcode"
DEFAULT_GEOG_COLS = [
    "pcon", "pct", "nuts", "lea", "statsward", "casward",
    "lsoa01", "msoa01", "ur01ind", "oac01", "lsoa11",
    "msoa11", "parish", "bua11", "buasd11", "ru11ind",
    "oac11", "imd",
]

# Columns added:
CRATE_COL_PK = "crate_pk"
# Do NOT use 'id'; this appears in RiO ClientAlternativeId /
# RCEP Client_Alternative_ID. "pk" is OK for RCEP + RiO, but clarity is good
CRATE_COL_RIO_NUMBER = "crate_rio_number"
# "rio_number" is OK for RCEP + RiO, but clarity is good
CRATE_COL_NHS_NUMBER = "crate_nhs_number_int"
# "nhs_number_int" is OK for RCEP + RiO, but again...
# For RCEP, in SQL Server, check existing columns with:
#   USE database;
#   SELECT column_name, table_name
#       FROM information_schema.columns
#       WHERE column_name = 'something';
# For RiO, for now, check against documented table structure.

# For progress notes:
CRATE_COL_MAX_SUBNUM = "crate_max_subnum_for_notenum"
CRATE_COL_LAST_NOTE = "crate_last_note_in_edit_chain"
# For clinical documents:
CRATE_COL_MAX_DOCVER = "crate_max_docver_for_doc"
CRATE_COL_LAST_DOC = "crate_last_doc_in_chain"

# Indexes added... generic:
CRATE_IDX_PK = "crate_idx_pk"  # for any patient table
CRATE_IDX_RIONUM = "crate_idx_rionum"  # for any patient table
# For progress notes:
CRATE_IDX_RIONUM_NOTENUM = "crate_idx_rionum_notenum"
CRATE_IDX_MAX_SUBNUM = "crate_idx_max_subnum"
CRATE_IDX_LAST_NOTE = "crate_idx_last_note"
# For clinical documents:
CRATE_IDX_RIONUM_SERIALNUM = "crateidx_rionum_serialnum"
CRATE_IDX_MAX_DOCVER = "crate_idx_max_docver"
CRATE_IDX_LAST_DOC = "crate_idx_last_doc"

# Views added:
VIEW_RCEP_CPFT_PROGRESS_NOTES_CURRENT = "progress_notes_current_crate"
VIEW_ADDRESS_WITH_GEOGRAPHY = "client_address_with_geography"



# -*- coding: utf-8 -*-
# Generated by Django 1.9.8 on 2017-02-06 16:17
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('userprofile', '0003_auto_20160628_1301'),
    ]

    operations = [
        migrations.AddField(
            model_name='userprofile',
            name='patients_per_page',
            field=models.PositiveSmallIntegerField(choices=[(1, '1'), (5, '5'), (10, '10'), (20, '20'), (50, '50'), (100, '100')], default=1, verbose_name='Number of patients to show per page (for Patient Explorer view)'),  # noqa
        ),
    ]



#!/usr/bin/env python
# crate_anon/crateweb/extra/django_cache_fn.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================

Based on https://github.com/rchrd2/django-cache-decorator
but fixed for Python 3 / Django 1.10.

"""

import hashlib
import logging
from typing import Any, Callable, Dict, Tuple

from django.core.cache import cache  # default cache

from crate_anon.common.jsonfunc import json_encode

FunctionType = Callable[..., Any]
ArgsType = Tuple[Any, ...]
KwargsType = Dict[str, Any]

log = logging.getLogger(__name__)
DEBUG_CACHE = False


def get_call_signature(fn: FunctionType,
                       args: ArgsType,
                       kwargs: KwargsType) -> str:
    # Note that the function won't have the __self__ argument (as in
    # fn.__self__), at this point, even if it's a member function.
    try:
        call_sig = json_encode((fn.__qualname__, args, kwargs))
    except TypeError:
        log.critical(
            "\nTo decorate using @django_cache_function without specifying "
            "cache_key, the decorated function's owning class and its "
            "parameters must be JSON-serializable (see jsonfunc.py, "
            "django_cache_fn.py).\n")
        raise
    if DEBUG_CACHE:
        log.debug("Making call signature {}".format(repr(call_sig)))
    return call_sig


def make_cache_key(call_signature: str) -> str:
    # - We have a bunch of components of arbitrary type, and we need to get
    #   a unique string out.
    # - We shouldn't use str(), because that is often poorly specified; e.g.
    #   is 'a.b.c' a TableId, or is it a ColumnId with no 'db' field?
    # - We could use repr(): sometimes that gives us helpful things that
    #   could in principle be passed to eval(), in which case repr() would
    #   be fine, but sometimes it doesn't, and gives unhelpful things like
    #   '<__main__.Thing object at 0x7ff3093ebda0>'.
    # - However, if something encodes to JSON, that representation should
    #   be reversible and thus contain the right sort of information.
    # - Note also that bound methods will come with a "self" argument, for
    #   which the address may be very relevant...
    # - Let's go with repr(). Users of the cache decorator should not pass
    #   objects whose repr() includes the memory address of the object unless
    #   they want those objects to be treated as distinct.
    # - Ah, no. The cache itself will pickle and unpickle things, and this
    #   will change memory addresses of objects. So we can't store a reference
    #   to an object using repr() and using cache.add()/pickle() and hope
    #   they'll come out the same.
    # - Use the JSON after all.
    # - And do it in get_call_signature(), not here.
    # - That means that any class we wish to decorate WITHOUT specifying a
    #   cache key manually must support JSON.
    key = hashlib.md5(call_signature.encode("utf-8")).hexdigest()
    if DEBUG_CACHE:
        log.debug("Making cache key {} from call_signature {}".format(
            key, repr(call_signature)))
    return key


def django_cache_function(timeout: int = 5 * 60,
                          cache_key: str = ''):
    """
    Decorator to add caching to a function in Django.
    Uses the Django default cache.
    Args:
        timeout: timeout in seconds; use None for "never expire", as 0 means
            "do not cache".
        cache_key: optional cache key to use (if falsy, we'll invent one)
    """
    cache_key = cache_key or None

    def decorator(fn):
        def wrapper(*args, **kwargs):
            # - NOTE that Django returns None from cache.get() for "not in
            #   cache", so can't cache a None value;
            #   https://docs.djangoproject.com/en/1.10/topics/cache/#basic-usage  # noqa
            # - We need to store a bit more than just the function result
            #   anyway, to detect hash collisions when the user doesn't specify
            #   the cache_key, so we may as well use that format even if the
            #   user does specify the cache_key, and then we can store a None
            #   result properly as well.
            if cache_key:
                # User specified a cache key. This is easy.
                call_sig = ''
                _cache_key = cache_key
                check_stored_call_sig = False
            else:
                # User didn't specify a cache key, so we'll do one
                # automatically. Since we do this via a hash, there is a small
                # but non-zero chance of a hash collision.
                call_sig = get_call_signature(fn, args, kwargs)
                _cache_key = make_cache_key(call_sig)
                check_stored_call_sig = True
            if DEBUG_CACHE:
                log.critical("Checking cache for key: " + _cache_key)
            cache_result_tuple = cache.get(_cache_key)  # TALKS TO CACHE HERE
            if cache_result_tuple is None:
                if DEBUG_CACHE:
                    log.critical("Cache miss")
            else:
                if DEBUG_CACHE:
                    log.critical("Cache hit")
                cached_call_sig, func_result = cache_result_tuple
                if (not check_stored_call_sig) or cached_call_sig == call_sig:
                    return func_result
                log.warning(
                    "... Cache hit was due to hash collision; cached_call_sig "
                    "{} != call_sig {}".format(
                        repr(cached_call_sig), repr(call_sig)))
                # If we get here, either it wasn't in the cache, or something
                # was in the cache that matched by cache_key but was actually a
                # hash collision. Either way, we must do the real work.
            func_result = fn(*args, **kwargs)
            cache_result_tuple = (call_sig, func_result)
            cache.set(key=_cache_key, value=cache_result_tuple,
                      timeout=timeout)  # TALKS TO CACHE HERE
            return func_result

        return wrapper

    return decorator



#!/usr/bin/env python
# crate_anon/nlp_manager/test_all_regex.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import argparse

from crate_anon.nlp_manager import (
    all_processors,
    regex_parser,
    regex_units,
)


def test_all_regex_nlp(verbose: bool = False) -> None:
    regex_parser.test_all(verbose=verbose)  # basic regexes
    regex_units.test_all(verbose=verbose)
    all_processors.test_all_processors(verbose=verbose)
    # ... tests all parser classes


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--verbose', '-v', action="store_true", help="Verbose")
    args = parser.parse_args()
    test_all_regex_nlp(verbose=args.verbose)



#!/usr/bin/env python
# crate_anon/anonymise/test_anonymisation.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================

Test the anonymisation for specific databases.

From the output, we have:
    n_replacements (POSITIVE)
    word_count (N)
    true_positive_confidential_masked (TP)
    false_positive_banal_masked (FP)
    false_negative_confidential_visible_known_to_source (FN)
    confidential_visible_but_unknown_to_source

Therefore, having summed across documents:
    TP + FP = POSITIVE
    NEGATIVE = N - POSITIVE
    TN = NEGATIVE - FN

and then we have everything we need. For all identifiers, we make FN equal to
    false_negative_confidential_visible_known_to_source
        + not_false_negative_confidential_visible_but_unknown_to_source
instead.
"""

# =============================================================================
# Imports
# =============================================================================

# from __future__ import print_function
import argparse
import collections
import csv
import json
import logging
import os
from typing import Any, Dict, List, Optional, Tuple

from crate_anon.anonymise.config_singleton import config
from crate_anon.anonymise.patient import Patient
from crate_anon.common.fileops import mkdir_p
from crate_anon.common.logsupport import configure_logger_for_colour

log = logging.getLogger(__name__)


# =============================================================================
# Specific tests
# =============================================================================

class FieldInfo(object):
    def __init__(self, table: str, field: str) -> None:
        """
        Fetches useful subsets from the data dictionary.
        """
        ddrows = config.dd.get_rows_for_dest_table(table)
        if not ddrows:
            raise ValueError("No data dictionary rows for destination table "
                             "{}".format(table))
        try:
            textrow = next(x for x in ddrows if x.dest_field == field)
        except StopIteration:
            raise ValueError("No destination field: {}".format(field))
        try:
            pkrow = next(x for x in ddrows if x.pk)
        except StopIteration:
            raise ValueError("No PK field found")
        try:
            pidrow = next(x for x in ddrows if x.primary_pid)
        except StopIteration:
            raise ValueError("No PID field found")

        self.pk_ddrow = pkrow
        self.pid_ddrow = pidrow
        self.text_ddrow = textrow
        log.info("Using fields: pk={}, pid={}, text={}".format(
            self.pk_ddrow.get_dest_signature(),
            self.pid_ddrow.get_dest_signature(),
            self.text_ddrow.get_dest_signature()))


def get_patientnum_rawtext(docid: int,
                           fieldinfo: FieldInfo) -> Tuple[Optional[int],
                                                          Optional[str]]:
    """
    Fetches the original text for a given document PK, plus the associated
    patient ID.
    """
    db = config.sources[fieldinfo.text_ddrow.src_db]
    table = fieldinfo.text_ddrow.src_table
    textfield = fieldinfo.text_ddrow.src_field
    sourcedbname = fieldinfo.text_ddrow.src_db
    pidfield = fieldinfo.pid_ddrow.src_field
    pkfield = fieldinfo.pk_ddrow.src_field
    src_ddrows = config.dd.get_rows_for_src_table(sourcedbname, table)
    sourcefields = []
    idx_pidfield = None
    idx_textfield = None
    for i, ddr in enumerate(src_ddrows):
        sourcefields.append(ddr.src_field)
        if ddr.src_field == pidfield:
            idx_pidfield = i
        if ddr.src_field == textfield:
            idx_textfield = i
    if idx_pidfield is None:
        raise ValueError("Unknown idx_pidfield")
    if idx_textfield is None:
        raise ValueError("Unknown idx_textfield")
    query = """
        SELECT {fields}
        FROM {table}
        WHERE {pkfield} = ?
    """.format(
        fields=",".join(sourcefields),
        textfield=textfield,
        table=table,
        pkfield=pkfield,
    )
    # log.debug("RAW: {}, {}".format(query, docid))
    row = db.fetchone(query, docid)
    if not row:
        return None, None
    pid = row[idx_pidfield]
    text = row[idx_textfield]
    ddr = src_ddrows[idx_textfield]
    for altermethod in fieldinfo.text_ddrow.get_extracting_text_altermethods():
        text, _ = altermethod.alter(config, text, ddr, row, src_ddrows)
    return pid, text


def get_patientnum_anontext(docid: int,
                            fieldinfo: FieldInfo) -> Tuple[Optional[int],
                                                           Optional[str]]:
    """
    Fetches the anonymised text for a given document PK, plus the associated
    patient ID.
    """
    db = config.destdb
    table = fieldinfo.text_ddrow.dest_table
    textfield = fieldinfo.text_ddrow.dest_field
    pidfield = fieldinfo.pid_ddrow.dest_field
    pkfield = fieldinfo.pk_ddrow.dest_field
    query = """
        SELECT {pidfield}, {textfield}
        FROM {table}
        WHERE {pkfield} = ?
    """.format(
        pidfield=pidfield,
        textfield=textfield,
        table=table,
        pkfield=pkfield,
    )
    # log.debug("ANON: {}, {}".format(query, docid))
    result = db.fetchone(query, docid)
    if not result:
        return None, None
    pid, text = result
    return pid, text


def process_doc(docid: int,
                args: Any,
                fieldinfo: FieldInfo,
                csvwriter: Any,
                first: bool,
                scrubdict: Dict[int, Dict[str, Any]]) -> int:
    """
    Write the original and anonymised documents to disk, plus some
    counts to a CSV file.
    """
    # Get stuff
    patientnum, rawtext = get_patientnum_rawtext(docid, fieldinfo)
    patientnum2, anontext = get_patientnum_anontext(docid, fieldinfo)
    # patientnum is raw; patientnum2 is hashed

    # Get scrubbing info
    patient = Patient(patientnum)
    scrubber = patient.scrubber
    scrubdict[patientnum] = scrubber.get_raw_info()

    # Write text
    rawfilename = os.path.join(args.rawdir,
                               "{}_{}.txt".format(patientnum, docid))
    anonfilename = os.path.join(args.anondir,
                                "{}_{}.txt".format(patientnum, docid))
    with open(rawfilename, 'w') as f:
        if rawtext:
            f.write(rawtext)
    with open(anonfilename, 'w') as f:
        if anontext:
            f.write(anontext)

    wordcount = len(rawtext.split()) if rawtext else 0

    if anontext:
        n_patient = anontext.count(config.replace_patient_info_with)
        n_thirdparty = anontext.count(config.replace_third_party_info_with)
        n_nonspecific = anontext.count(config.replace_nonspecific_info_with)
    else:
        n_patient = 0
        n_thirdparty = 0
        n_nonspecific = 0
    n_replacements = n_patient + n_thirdparty + n_nonspecific

    summary = collections.OrderedDict()
    summary["src_db"] = fieldinfo.text_ddrow.src_db
    summary["src_table"] = fieldinfo.text_ddrow.src_table
    summary["src_field_pid"] = fieldinfo.pid_ddrow.src_field
    summary["pid"] = patientnum
    summary["src_field_pk"] = fieldinfo.pk_ddrow.src_field
    summary["docid"] = docid
    summary["src_field_text"] = fieldinfo.text_ddrow.src_field
    summary["dest_table"] = fieldinfo.text_ddrow.dest_table
    summary["dest_field"] = fieldinfo.text_ddrow.dest_field
    summary["n_replacements"] = n_replacements
    # summary["n_patient"] = n_patient
    # summary["n_thirdparty"] = n_thirdparty
    # summary["n_nonspecific"] = n_nonspecific
    summary["word_count"] = wordcount
    # ... use this to calculate true negatives (banal, visible) as:
    # true_negative = word_count - (true_pos + false_pos + false_neg)
    summary["true_positive_confidential_masked"] = "?"
    summary["false_positive_banal_masked"] = "?"
    summary["false_negative_confidential_visible_known_to_source"] = "?"
    summary["confidential_visible_but_unknown_to_source"] = "?"
    summary["comments"] = ""

    if first:
        csvwriter.writerow(list(summary.keys()))
    csvwriter.writerow(list(summary.values()))

    return patientnum


def get_docids(args: Any,
               fieldinfo: FieldInfo,
               from_src: bool = True) -> List[int]:
    """
    Generate a limited set of PKs for the documents.
    """
    if from_src:
        db = config.sources[fieldinfo.text_ddrow.src_db]
        table = fieldinfo.pk_ddrow.src_table
        pkfield = fieldinfo.pk_ddrow.src_field
        pidfield = fieldinfo.pid_ddrow.src_field
    else:
        db = config.destdb
        table = fieldinfo.pk_ddrow.dest_table
        pkfield = fieldinfo.pk_ddrow.dest_field
        pidfield = fieldinfo.pid_ddrow.dest_field
    if args.uniquepatients:
        query = """
            SELECT MIN({pkfield}), {pidfield}
            FROM {table}
            GROUP BY {pidfield}
            ORDER BY {pidfield}
            LIMIT {limit}
        """.format(
            pkfield=pkfield,
            pidfield=pidfield,
            table=table,
            limit=args.limit,
        )
        return db.fetchallfirstvalues(query)
    else:
        query = """
            SELECT {pkfield}
            FROM {table}
            ORDER BY {pkfield}
            LIMIT {limit}
        """.format(
            pkfield=pkfield,
            table=table,
            limit=args.limit,
        )
        return db.fetchallfirstvalues(query)


def test_anon(args: Any) -> None:
    """
    Fetch raw and anonymised documents and store them in files for
    comparison, along with some summary information.
    """
    fieldinfo = FieldInfo(args.dsttable, args.dstfield)
    docids = get_docids(args, fieldinfo, args.from_src)
    mkdir_p(args.rawdir)
    mkdir_p(args.anondir)
    scrubdict = {}
    pidset = set()
    with open(args.resultsfile, 'w') as csvfile:
        csvwriter = csv.writer(csvfile, delimiter='\t')
        first = True
        for docid in docids:
            # noinspection PyTypeChecker
            pid = process_doc(docid, args, fieldinfo, csvwriter, first,
                              scrubdict)
            first = False
            pidset.add(pid)
    with open(args.scrubfile, 'w') as f:
        f.write(json.dumps(scrubdict, indent=4))
    log.info("Finished. See {} for a summary.".format(args.resultsfile))
    log.info(
        "Use meld to compare directories {} and {}".format(
            args.rawdir,
            args.anondir,
        )
    )
    log.info("To install meld on Debian/Ubuntu: sudo apt-get install meld")
    log.info("{} documents, {} patients".format(len(docids), len(pidset)))


# =============================================================================
# Main
# =============================================================================

def main() -> None:
    """
    Command-line entry point.
    """
    parser = argparse.ArgumentParser(
        description='Test anonymisation',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--config', required=True,
                        help='Configuration file name (input)')
    parser.add_argument('--dsttable', required=True,
                        help='Destination table')
    parser.add_argument('--dstfield', required=True,
                        help='Destination column')
    parser.add_argument('--limit', type=int, default=100,
                        help='Limit on number of documents')
    parser.add_argument('--rawdir', default='raw',
                        help='Directory for raw output text files')
    parser.add_argument('--anondir', default='anon',
                        help='Directory for anonymised output text files')
    parser.add_argument('--resultsfile', default='testanon_results.csv',
                        help='Results output CSV file name')
    parser.add_argument('--scrubfile', default='testanon_scrubber.txt',
                        help='Scrubbing information text file name')
    parser.add_argument('--verbose', '-v', action='store_true',
                        help="Be verbose")

    pkgroup = parser.add_mutually_exclusive_group(required=False)
    pkgroup.add_argument('--pkfromsrc', dest='from_src', action='store_true',
                         help='Fetch PKs (document IDs) from source (default)')
    pkgroup.add_argument('--pkfromdest', dest='from_src', action='store_false',
                         help='Fetch PKs (document IDs) from destination')
    parser.set_defaults(from_src=True)

    uniquegroup = parser.add_mutually_exclusive_group(required=False)
    uniquegroup.add_argument(
        '--uniquepatients', dest='uniquepatients', action='store_true',
        help='Only one document per patient (the first by PK) (default)')
    uniquegroup.add_argument(
        '--nonuniquepatients', dest='uniquepatients', action='store_false',
        help='Documents in sequence, with potentially >1 document/patient')
    parser.set_defaults(uniquepatients=True)

    args = parser.parse_args()

    loglevel = logging.DEBUG if args.verbose else logging.INFO
    rootlogger = logging.getLogger()
    configure_logger_for_colour(rootlogger, loglevel)

    log.info("Arguments: " + str(args))

    # Load/validate config
    log.info("Loading config...")
    config.set(filename=args.config, load_destfields=False)
    log.info("... config loaded")

    # Do it
    test_anon(args)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# crate_anon/common/sqla.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import ast
import contextlib
import copy
from functools import lru_cache
import logging
import re
from typing import Any, Dict, List, Optional, Tuple, Type, Union

from sqlalchemy.dialects import mssql, mysql
# noinspection PyUnresolvedReferences
from sqlalchemy.engine import Connection, Engine, ResultProxy
from sqlalchemy.engine.interfaces import Dialect
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.ext.declarative.api import DeclarativeMeta
from sqlalchemy.schema import (Column, CreateColumn, DDL, MetaData, Index,
                               Sequence, Table)
from sqlalchemy.orm.session import Session
from sqlalchemy.sql import (column, exists, func, literal, select, sqltypes,
                            text, table)
from sqlalchemy.sql.expression import (
    ClauseElement,
    Insert,
    TableClause,
)
from sqlalchemy.sql.sqltypes import BigInteger, TypeEngine

log = logging.getLogger(__name__)


# =============================================================================
# Constants
# =============================================================================

MSSQL_DEFAULT_SCHEMA = 'dbo'
POSTGRES_DEFAULT_SCHEMA = 'public'


# =============================================================================
# SQL - SQLAlchemy ORM
# =============================================================================

# -----------------------------------------------------------------------------
# Get or create (SQLAlchemy ORM)
# -----------------------------------------------------------------------------
# http://stackoverflow.com/questions/2546207
# ... composite of several suggestions

def get_or_create(session: Session,
                  model: DeclarativeMeta,
                  defaults: Dict[str, Any] = None,
                  **kwargs: Any) -> Tuple[Any, bool]:
    instance = session.query(model).filter_by(**kwargs).first()
    if instance:
        return instance, False
    else:
        params = dict((k, v) for k, v in kwargs.items()
                      if not isinstance(v, ClauseElement))
        params.update(defaults or {})
        instance = model(**params)
        session.add(instance)
        return instance, True


# =============================================================================
# SQL - SQLAlchemy Core
# =============================================================================

# -----------------------------------------------------------------------------
# SELECT COUNT(*) (SQLAlchemy Core)
# -----------------------------------------------------------------------------
# http://stackoverflow.com/questions/12941416

def count_star(session: Union[Session, Engine, Connection],
               tablename: str) -> int:
    # works if you pass a connection or a session or an engine; all have
    # the execute() method
    query = select([func.count()]).select_from(table(tablename))
    return session.execute(query).scalar()


# -----------------------------------------------------------------------------
# SELECT COUNT(*), MAX(field) (SQLAlchemy Core)
# -----------------------------------------------------------------------------

def count_star_and_max(session: Union[Session, Engine, Connection],
                       tablename: str,
                       maxfield: str) -> Tuple[int, Optional[int]]:
    query = select([
        func.count(),
        func.max(column(maxfield))
    ]).select_from(table(tablename))
    result = session.execute(query)
    return result.fetchone()  # count, maximum


# -----------------------------------------------------------------------------
# SELECT EXISTS (SQLAlchemy Core)
# -----------------------------------------------------------------------------
# http://stackoverflow.com/questions/15381604
# http://docs.sqlalchemy.org/en/latest/orm/query.html


def exists_plain(session: Session, tablename: str, *criteria: Any) -> bool:
    exists_clause = exists().select_from(table(tablename))
    # ... EXISTS (SELECT * FROM tablename)
    for criterion in criteria:
        exists_clause = exists_clause.where(criterion)
    # ... EXISTS (SELECT * FROM tablename WHERE ...)

    if session.get_bind().dialect.name == 'mssql':
        query = select([literal(True)]).where(exists_clause)
        # ... SELECT 1 WHERE EXISTS (SELECT * FROM tablename WHERE ...)
    else:
        query = select([exists_clause])
        # ... SELECT EXISTS (SELECT * FROM tablename WHERE ...)

    result = session.execute(query).scalar()
    return bool(result)


def exists_orm(session: Session,
               ormclass: DeclarativeMeta,
               *criteria: Any) -> bool:
    # http://docs.sqlalchemy.org/en/latest/orm/query.html
    q = session.query(ormclass)
    for criterion in criteria:
        q = q.filter(criterion)

    # See this:
    # - https://bitbucket.org/zzzeek/sqlalchemy/issues/3212/misleading-documentation-for-queryexists  # noqa
    # - http://docs.sqlalchemy.org/en/latest/orm/query.html#sqlalchemy.orm.query.Query.exists  # noqa

    exists_clause = q.exists()
    if session.get_bind().dialect.name == 'mssql':
        # SQL Server
        result = session.query(literal(True)).filter(exists_clause).scalar()
        # SELECT 1 WHERE EXISTS (SELECT 1 FROM table WHERE ...)
        # ... giving 1 or None (no rows)
        # ... fine for SQL Server, but invalid for MySQL (no FROM clause)
    else:
        # MySQL, etc.
        result = session.query(exists_clause).scalar()
        # SELECT EXISTS (SELECT 1 FROM table WHERE ...)
        # ... giving 1 or 0
        # ... fine for MySQL, but invalid syntax for SQL server
    return bool(result)


# =============================================================================
# INSERT ... ON DUPLICATE KEY UPDATE support, for MySQL
# =============================================================================
# https://www.reddit.com/r/Python/comments/p5grh/sqlalchemy_whats_the_idiomatic_way_of_writing/  # noqa
# https://github.com/bedwards/sqlalchemy_mysql_ext/blob/master/duplicate.py
# ... modified
# http://docs.sqlalchemy.org/en/rel_1_0/core/compiler.html
# http://stackoverflow.com/questions/6611563/sqlalchemy-on-duplicate-key-update
# http://dev.mysql.com/doc/refman/5.7/en/insert-on-duplicate.html

# noinspection PyAbstractClass
class InsertOnDuplicate(Insert):
    pass


def insert_on_duplicate(tablename, values=None, inline=False, **kwargs):
    return InsertOnDuplicate(tablename, values, inline=inline, **kwargs)


# noinspection PyPep8Naming
def monkeypatch_TableClause():
    log.debug("Adding 'INSERT ON DUPLICATE KEY UPDATE' support for MySQL "
              "to SQLAlchemy")
    TableClause.insert_on_duplicate = insert_on_duplicate


# noinspection PyPep8Naming
def unmonkeypatch_TableClause():
    del TableClause.insert_on_duplicate


STARTSEPS = '`'
ENDSEPS = '`'
INSERT_FIELDNAMES_REGEX = (
    r'^INSERT\sINTO\s[{startseps}]?(?P<table>\w+)[{endseps}]?\s+'
    r'\((?P<columns>[, {startseps}{endseps}\w]+)\)\s+VALUES'.format(
        startseps=STARTSEPS, endseps=ENDSEPS)
)
# http://pythex.org/ !
RE_INSERT_FIELDNAMES = re.compile(INSERT_FIELDNAMES_REGEX)


@compiles(InsertOnDuplicate, 'mysql')
def compile_insert_on_duplicate_key_update(insert, compiler, **kw):
    """
    We can't get the fieldnames directly from 'insert' or 'compiler'.
    We could rewrite the innards of the visit_insert statement, like
        https://github.com/bedwards/sqlalchemy_mysql_ext/blob/master/duplicate.py  # noqa
    ... but, like that, it will get outdated.
    We could use a hack-in-by-hand method, like
        http://stackoverflow.com/questions/6611563/sqlalchemy-on-duplicate-key-update
    ... but a little automation would be nice.
    So, regex to the rescue:
    """
    # log.critical(compiler.__dict__)
    # log.critical(compiler.dialect.__dict__)
    # log.critical(insert.__dict__)
    s = compiler.visit_insert(insert, **kw)
    # log.critical(s)
    m = RE_INSERT_FIELDNAMES.match(s)
    if m is None:
        raise ValueError("compile_insert_on_duplicate_key_update: no match")
    columns = [c.strip() for c in m.group('columns').split(",")]
    # log.critical(columns)
    updates = ", ".join(["{c} = VALUES({c})".format(c=c) for c in columns])
    s += ' ON DUPLICATE KEY UPDATE {}'.format(updates)
    # log.critical(s)
    return s


# =============================================================================
# Inspect tables (SQLAlchemy Core)
# =============================================================================

def get_table_names(engine: Engine) -> List[str]:
    insp = Inspector.from_engine(engine)
    return insp.get_table_names()


def table_exists(engine: Engine, tablename: str) -> bool:
    return tablename in get_table_names(engine)


def get_columns_info(engine: Engine, tablename: str) -> List[Dict]:
    insp = Inspector.from_engine(engine)
    return insp.get_columns(tablename)


def get_column_info(engine: Engine, tablename: str,
                    columnname: str) -> Optional[Dict]:
    # Dictionary structure: see
    # http://docs.sqlalchemy.org/en/latest/core/reflection.html#sqlalchemy.engine.reflection.Inspector.get_columns  # noqa
    columns = get_columns_info(engine, tablename)
    for x in columns:
        if x['name'] == columnname:
            return x
    return None


def get_column_type(engine: Engine, tablename: str,
                    columnname: str) -> Optional[TypeEngine]:
    # Dictionary structure: see
    # http://docs.sqlalchemy.org/en/latest/core/reflection.html#sqlalchemy.engine.reflection.Inspector.get_columns  # noqa
    columns = get_columns_info(engine, tablename)
    for x in columns:
        if x['name'] == columnname:
            return x['type']
    return None


def get_column_names(engine: Engine, tablename: str) -> List[str]:
    return [x['name'] for x in get_columns_info(engine, tablename)]


# =============================================================================
# More introspection
# =============================================================================

def get_single_int_pk_colname(table_: Table) -> Optional[str]:
    """
    If a table has a single-field (non-composite) integer PK, this will
    return its name; otherwise, None.

    Note that it is fine to have both a composite primary key and a separate
    IDENTITY (AUTOINCREMENT) integer field.
    """
    n_pks = 0
    int_pk_names = []
    for col in table_.columns:
        if col.primary_key:
            n_pks += 1
            if is_sqlatype_integer(col.type):
                int_pk_names.append(col.name)
    if n_pks == 1 and len(int_pk_names) == 1:
        return int_pk_names[0]
    return None


def get_single_int_autoincrement_colname(table_: Table) -> Optional[str]:
    """
    If a table has a single integer AUTOINCREMENT column, this will
    return its name; otherwise, None.

    - It's unlikely that a database has >1 AUTOINCREMENT field anyway, but we
      should check.
    - SQL Server's IDENTITY keyword is equivalent to MySQL's AUTOINCREMENT.
    - Verify against SQL Server:

        SELECT table_name, column_name
        FROM information_schema.columns
        WHERE COLUMNPROPERTY(OBJECT_ID(table_schema + '.' + table_name),
                             column_name,
                             'IsIdentity') = 1
        ORDER BY table_name;

      ... http://stackoverflow.com/questions/87747

    - Also:

        sp_columns 'tablename';

        ... which is what SQLAlchemy does (dialects/mssql/base.py, in
        get_columns).
    """
    n_autoinc = 0
    int_autoinc_names = []
    for col in table_.columns:
        if col.autoincrement:
            n_autoinc += 1
            if is_sqlatype_integer(col.type):
                int_autoinc_names.append(col.name)
    if n_autoinc > 1:
        log.warning("Table {} has {} autoincrement columns".format(
            repr(table_.name), n_autoinc))
    if n_autoinc == 1 and len(int_autoinc_names) == 1:
        return int_autoinc_names[0]
    return None


def get_effective_int_pk_col(table_: Table) -> Optional[str]:
    return (
        get_single_int_pk_colname(table_) or
        get_single_int_autoincrement_colname(table_) or
        None
    )


# =============================================================================
# Indexes
# =============================================================================

def index_exists(engine: Engine, tablename: str, indexname: str) -> bool:
    insp = Inspector.from_engine(engine)
    return any(i['name'] == indexname for i in insp.get_indexes(tablename))


def mssql_get_pk_index_name(engine: Engine,
                            tablename: str,
                            schemaname: str = MSSQL_DEFAULT_SCHEMA) -> str:
    # http://docs.sqlalchemy.org/en/latest/core/connections.html#sqlalchemy.engine.Connection.execute  # noqa
    # http://docs.sqlalchemy.org/en/latest/core/sqlelement.html#sqlalchemy.sql.expression.text  # noqa
    # http://docs.sqlalchemy.org/en/latest/core/sqlelement.html#sqlalchemy.sql.expression.TextClause.bindparams  # noqa
    # http://docs.sqlalchemy.org/en/latest/core/connections.html#sqlalchemy.engine.ResultProxy  # noqa
    query = text("""
SELECT
    kc.name AS index_name
FROM
    sys.key_constraints AS kc
    INNER JOIN sys.tables AS ta ON ta.object_id = kc.parent_object_id
    INNER JOIN sys.schemas AS s ON ta.schema_id = s.schema_id
WHERE
    kc.[type] = 'PK'
    AND ta.name = :tablename
    AND s.name = :schemaname
    """).bindparams(
        tablename=tablename,
        schemaname=schemaname,
    )
    with contextlib.closing(engine.execute(query)) as result:  # type: ResultProxy  # noqa
        row = result.fetchone()
        return row[0] if row else ''


def mssql_table_has_ft_index(engine: Engine,
                             tablename: str,
                             schemaname: str = MSSQL_DEFAULT_SCHEMA) -> bool:
    query = text("""
SELECT
    COUNT(*)
FROM
    sys.key_constraints AS kc
    INNER JOIN sys.tables AS ta ON ta.object_id = kc.parent_object_id
    INNER JOIN sys.schemas AS s ON ta.schema_id = s.schema_id
    INNER JOIN sys.fulltext_indexes AS fi ON fi.object_id = ta.object_id
WHERE
    ta.name = :tablename
    AND s.name = :schemaname
    """).bindparams(
        tablename=tablename,
        schemaname=schemaname,
    )
    with contextlib.closing(engine.execute(query)) as result:  # type: ResultProxy  # noqa
        row = result.fetchone()
        return row[0] > 0


def mssql_transaction_count(engine_or_conn: Union[Connection, Engine]) -> int:
    sql = "SELECT @@TRANCOUNT"
    with contextlib.closing(engine_or_conn.execute(sql)) as result:  # type: ResultProxy  # noqa
        row = result.fetchone()
        return row[0] if row else None


def add_index(engine: Engine,
              sqla_column: Column = None,
              multiple_sqla_columns: List[Column] = None,
              unique: bool = False,
              fulltext: bool = False,
              length: int = None) -> None:
    # We used to process a table as a unit; this makes index creation faster
    # (using ALTER TABLE).
    # http://dev.mysql.com/doc/innodb/1.1/en/innodb-create-index-examples.html  # noqa
    # ... ignored in transition to SQLAlchemy
    is_mssql = engine.dialect.name == 'mssql'
    is_mysql = engine.dialect.name == 'mysql'

    multiple_sqla_columns = multiple_sqla_columns or []
    if multiple_sqla_columns and not (fulltext and is_mssql):
        raise ValueError("add_index: Use multiple_sqla_columns only for mssql "
                         "(Microsoft SQL Server) full-text indexing")
    if bool(multiple_sqla_columns) == (sqla_column is not None):
        raise ValueError(
            "add_index: Use either sqla_column or multiple_sqla_columns, not "
            "both (sqla_column = {}, multiple_sqla_columns = {}".format(
                repr(sqla_column), repr(multiple_sqla_columns)))
    if sqla_column is not None:
        colname = sqla_column.name
        sqla_table = sqla_column.table
        tablename = sqla_table.name
    else:
        colname = ", ".join(c.name for c in multiple_sqla_columns)
        sqla_table = multiple_sqla_columns[0].table
        tablename = sqla_table.name
        if any(c.table.name != tablename for c in multiple_sqla_columns[1:]):
            raise ValueError(
                "add_index: tablenames are inconsistent in "
                "multiple_sqla_columns = {}".format(
                    repr(multiple_sqla_columns)))

    if fulltext:
        if is_mssql:
            idxname = ''  # they are unnamed
        else:
            idxname = "_idxft_{}".format(colname)
    else:
        idxname = "_idx_{}".format(colname)
    if idxname and index_exists(engine, tablename, idxname):
        log.info("Skipping creation of index {} on table {}; already "
                 "exists".format(idxname, tablename))
        return
        # because it will crash if you add it again!
    log.info("Creating{ft} index {i} on table {t}, column {c}".format(
        ft=" full-text" if fulltext else "",
        i=idxname or "<unnamed>",
        t=tablename,
        c=colname))

    if fulltext:
        if is_mysql:
            log.warning('OK to ignore this warning: '
                        '"InnoDB rebuilding table to add column FTS_DOC_ID"')
            # https://dev.mysql.com/doc/refman/5.6/en/innodb-fulltext-index.html
            sql = (
                "ALTER TABLE {tablename} "
                "ADD FULLTEXT INDEX {idxname} ({colname})".format(
                    tablename=tablename,
                    idxname=idxname,
                    colname=colname,
                )
            )
            # DDL(sql, bind=engine).execute_if(dialect='mysql')
            DDL(sql, bind=engine).execute()

        elif is_mssql:  # Microsoft SQL Server
            # https://msdn.microsoft.com/library/ms187317(SQL.130).aspx
            # Argh! Complex.
            # Note that the database must also have had a
            #   CREATE FULLTEXT CATALOG somename AS DEFAULT;
            # statement executed on it beforehand.
            schemaname = engine.schema_for_object(sqla_table) or MSSQL_DEFAULT_SCHEMA  # noqa
            if mssql_table_has_ft_index(engine=engine,
                                        tablename=tablename,
                                        schemaname=schemaname):
                log.info(
                    "... skipping creation of full-text index on table {}; a "
                    "full-text index already exists for that table; you can "
                    "have only one full-text index per table, though it can "
                    "be on multiple columns".format(tablename))
                return
            pk_index_name = mssql_get_pk_index_name(
                engine=engine, tablename=tablename, schemaname=schemaname)
            if not pk_index_name:
                raise ValueError(
                    "To make a FULLTEXT index under SQL Server, we need to "
                    "know the name of the PK index, but couldn't find one "
                    "from get_pk_index_name() for table {}".format(
                        repr(tablename)))
            # We don't name the FULLTEXT index itself, but it has to relate
            # to an existing unique index.
            sql = (
                "CREATE FULLTEXT INDEX ON {tablename} ({colname}) "
                "KEY INDEX {keyidxname} ".format(
                    tablename=tablename,
                    keyidxname=pk_index_name,
                    colname=colname,
                )
            )
            # SQL Server won't let you do this inside a transaction:
            # "CREATE FULLTEXT INDEX statement cannot be used inside a user
            # transaction."
            # https://msdn.microsoft.com/nl-nl/library/ms191544(v=sql.105).aspx
            # So let's ensure any preceding transactions are completed, and
            # run the SQL in a raw way:
            # engine.execute(sql).execution_options(autocommit=False)
            # http://docs.sqlalchemy.org/en/latest/core/connections.html#understanding-autocommit
            #
            # ... lots of faff with this (see test code in no_transactions.py)
            # ... ended up using explicit "autocommit=True" parameter (for
            #     pyodbc); see create_indexes()
            transaction_count = mssql_transaction_count(engine)
            if transaction_count != 0:
                log.critical("SQL Server transaction count (should be 0): "
                             "{}".format(transaction_count))
                # Executing serial COMMITs or a ROLLBACK won't help here if
                # this transaction is due to Python DBAPI default behaviour.
            DDL(sql, bind=engine).execute()

            # The reversal procedure is DROP FULLTEXT INDEX ON tablename;

        else:
            log.error("Don't know how to make full text index on dialect "
                      "{}".format(engine.dialect.name))

    else:
        index = Index(idxname, sqla_column, unique=unique, mysql_length=length)
        index.create(engine)
    # Index creation doesn't require a commit.


# =============================================================================
# More DDL
# =============================================================================

def make_bigint_autoincrement_column(column_name: str,
                                     dialect: Dialect) -> Column:
    # noinspection PyUnresolvedReferences
    if dialect.name == 'mssql':
        return Column(column_name, BigInteger,
                      Sequence('dummy_name', start=1, increment=1))
    else:
        # return Column(column_name, BigInteger, autoincrement=True)
        # noinspection PyUnresolvedReferences
        raise AssertionError(
            "SQLAlchemy doesn't support non-PK autoincrement fields yet for "
            "dialect {}".format(repr(dialect.name)))
        # see http://stackoverflow.com/questions/2937229


def column_creation_ddl(sqla_column: Column, dialect: Dialect) -> str:
    """
    The column should already be bound to a table (because e.g. the SQL Server
    dialect requires this for DDL generation).

    Manual testing:

        from sqlalchemy.schema import Column, CreateColumn, MetaData, Sequence, Table
        from sqlalchemy.sql.sqltypes import BigInteger
        from sqlalchemy.dialects.mssql.base import MSDialect
        dialect = MSDialect()
        col1 = Column('hello', BigInteger, nullable=True)
        col2 = Column('world', BigInteger, autoincrement=True)  # does NOT generate IDENTITY
        col3 = Column('you', BigInteger, Sequence('dummy_name', start=1, increment=1))
        metadata = MetaData()
        t = Table('mytable', metadata)
        t.append_column(col1)
        t.append_column(col2)
        t.append_column(col3)
        print(str(CreateColumn(col1).compile(dialect=dialect)))  # hello BIGINT NULL
        print(str(CreateColumn(col2).compile(dialect=dialect)))  # world BIGINT NULL
        print(str(CreateColumn(col3).compile(dialect=dialect)))  # you BIGINT NOT NULL IDENTITY(1,1)

    If you don't append the column to a Table object, the DDL generation step
    gives:

        sqlalchemy.exc.CompileError: mssql requires Table-bound columns in
        order to generate DDL
    """  # noqa
    return str(CreateColumn(sqla_column).compile(dialect=dialect))


# noinspection PyUnresolvedReferences
def giant_text_sqltype(dialect: Dialect) -> str:
    """
    Args:
        dialect: a SQLAlchemy dialect class
    Returns:
        the SQL data type of "giant text", typically 'LONGTEXT' for MySQL
        and 'NVARCHAR(MAX)' for SQL Server.
    """
    if dialect.name == 'mssql':
        return 'NVARCHAR(MAX)'
    elif dialect.name == 'mysql':
        return 'LONGTEXT'
    else:
        raise ValueError("Unknown dialect: {}".format(dialect.name))


# =============================================================================
# SQLAlchemy column types
# =============================================================================

# -----------------------------------------------------------------------------
# Reverse a textual SQL column type to an SQLAlchemy column type
# -----------------------------------------------------------------------------

RE_COLTYPE_WITH_COLLATE = re.compile(r'(?P<maintype>.+) COLLATE .+')
RE_COLTYPE_WITH_ONE_PARAM = re.compile(r'(?P<type>\w+)\((?P<size>\w+)\)')
RE_COLTYPE_WITH_TWO_PARAMS = re.compile(
    r'(?P<type>\w+)\((?P<size>\w+),\s*(?P<dp>\w+)\)')
# http://www.w3schools.com/sql/sql_create_table.asp


def _get_sqla_coltype_class_from_str(coltype: str,
                                     dialect: Dialect) -> Type[Column]:
    """
    Upper- and lower-case search.
    For example, the SQLite dialect uses upper case, and the
    MySQL dialect uses lower case.
    """
    # noinspection PyUnresolvedReferences
    ischema_names = dialect.ischema_names
    try:
        return ischema_names[coltype.upper()]
    except KeyError:
        return ischema_names[coltype.lower()]


@lru_cache(maxsize=None)
def get_sqla_coltype_from_dialect_str(coltype: str,
                                      dialect: Dialect) -> TypeEngine:
    """
    Args:
        dialect: a SQLAlchemy dialect class
        coltype: a str() representation, e.g. from str(c['type']) where
            c is an instance of sqlalchemy.sql.schema.Column.
    Returns:
        a Python object that is a subclass of sqlalchemy.types.TypeEngine
    Example:
        get_sqla_coltype_from_string('INTEGER(11)', engine.dialect)
            -> Integer(length=11)

    Notes:
        -   sqlalchemy.engine.default.DefaultDialect is the dialect base class
        -   a dialect contains these things of interest:
                ischema_names: string-to-class dictionary
                type_compiler: instance of e.g.
                    sqlalchemy.sql.compiler.GenericTypeCompiler
                    ... has a process() method
                    ... but that operates on TypeEngine objects
                get_columns: takes a table name, inspects the database
        -   example of the dangers of eval:
            http://nedbatchelder.com/blog/201206/eval_really_is_dangerous.html
        -   An example of a function doing the reflection/inspection within
            SQLAlchemy is sqlalchemy.dialects.mssql.base.MSDialect.get_columns,
            which has this lookup:
                coltype = self.ischema_names.get(type, None)

    Caveats:
        -   the parameters, e.g. DATETIME(6), do NOT necessarily either work at
            all or work correctly. For example, SQLAlchemy will happily spit
            out 'INTEGER(11)' but its sqlalchemy.sql.sqltypes.INTEGER class
            takes no parameters, so you get:
                TypeError: object() takes no parameters
            Similarly, MySQL's DATETIME(6) uses the 6 to refer to precision,
            but the DATETIME class in SQLAlchemy takes only a boolean parameter
            (timezone).
        -   However, sometimes we have to have parameters, e.g. VARCHAR length.
        -   Thus, this is a bit useless.
        -   Fixed, with a few special cases.
    """
    size = None
    dp = None
    args = []
    kwargs = {}

    try:
        # Split e.g. "VARCHAR(32) COLLATE blah" into "VARCHAR(32)", "who cares"
        m = RE_COLTYPE_WITH_COLLATE.match(coltype)
        if m is not None:
            coltype = m.group('maintype')

        # Split e.g. "DECIMAL(10, 2)" into DECIMAL, 10, 2
        m = RE_COLTYPE_WITH_TWO_PARAMS.match(coltype)
        if m is not None:
            basetype = m.group('type').upper()
            size = ast.literal_eval(m.group('size'))
            dp = ast.literal_eval(m.group('dp'))
        else:
            # Split e.g. "VARCHAR(32)" into VARCHAR, 32
            m = RE_COLTYPE_WITH_ONE_PARAM.match(coltype)
            if m is not None:
                basetype = m.group('type').upper()
                size_text = m.group('size').strip().upper()
                if size_text != 'MAX':
                    size = ast.literal_eval(size_text)
            else:
                basetype = coltype.upper()

        # Special cases: pre-processing
        # noinspection PyUnresolvedReferences
        if dialect.name == 'mssql' and basetype.lower() == 'integer':
            basetype = 'int'

        cls = _get_sqla_coltype_class_from_str(basetype, dialect)

        # Special cases: post-processing
        if basetype == 'DATETIME' and size:
            # First argument to DATETIME() is timezone, so...
            # noinspection PyUnresolvedReferences
            if dialect.name == 'mysql':
                kwargs = {'fsp': size}
            else:
                pass
        else:
            args = [x for x in (size, dp) if x is not None]

        try:
            return cls(*args, **kwargs)
        except TypeError:
            return cls()

    except:
        # noinspection PyUnresolvedReferences
        raise ValueError("Failed to convert SQL type {} in dialect {} to an "
                         "SQLAlchemy type".format(repr(coltype),
                                                  repr(dialect.name)))

# get_sqla_coltype_from_dialect_str("INTEGER", engine.dialect)
# get_sqla_coltype_from_dialect_str("INTEGER(11)", engine.dialect)
# get_sqla_coltype_from_dialect_str("VARCHAR(50)", engine.dialect)
# get_sqla_coltype_from_dialect_str("DATETIME", engine.dialect)
# get_sqla_coltype_from_dialect_str("DATETIME(6)", engine.dialect)


# =============================================================================
# Do special dialect conversions on SQLAlchemy SQL types (of class type)
# =============================================================================

def remove_collation(coltype: TypeEngine) -> TypeEngine:
    if not getattr(coltype, 'collation', None):
        return coltype
    newcoltype = copy.copy(coltype)
    newcoltype.collation = None
    return newcoltype


@lru_cache(maxsize=None)
def convert_sqla_type_for_dialect(
        coltype: TypeEngine,
        dialect: Dialect,
        strip_collation: bool = True,
        convert_mssql_timestamp: bool = True,
        expand_for_scrubbing: bool = False) -> TypeEngine:
    """
    - The purpose of expand_for_scrubbing is that, for example, a VARCHAR(200)
      field containing one or more instances of "Jones", where "Jones" is to be
      replaced with "[XXXXXX]", will get longer (by an unpredictable amount).
      So, better to expand to unlimited length.
    """
    # log.critical("Incoming coltype: {}, vars={}".format(repr(coltype),
    #                                                     vars(coltype)))
    # noinspection PyUnresolvedReferences
    to_mysql = dialect.name == 'mysql'
    # noinspection PyUnresolvedReferences
    to_mssql = dialect.name == 'mssql'
    typeclass = type(coltype)
    
    # -------------------------------------------------------------------------
    # Text
    # -------------------------------------------------------------------------
    if isinstance(coltype, sqltypes.UnicodeText):
        # Unbounded Unicode text.
        # Includes derived classes such as mssql.base.NTEXT.
        return sqltypes.UnicodeText()
    if isinstance(coltype, sqltypes.Text):
        # Unbounded text, more generally. (UnicodeText inherits from Text.)
        # Includes sqltypes.TEXT.
        return sqltypes.Text()
    # Everything inheriting from String has a length property, but can be None.
    # There are types that can be unlimited in SQL Server, e.g. VARCHAR(MAX)
    # and NVARCHAR(MAX), that MySQL needs a length for. (Failure to convert
    # gives e.g.: 'NVARCHAR requires a length on dialect mysql'.)
    if isinstance(coltype, sqltypes.Unicode):
        # Includes NVARCHAR(MAX) in SQL -> NVARCHAR() in SQLAlchemy.
        if (coltype.length is None and to_mysql) or expand_for_scrubbing:
            return sqltypes.UnicodeText()
    # The most general case; will pick up any other string types.
    if isinstance(coltype, sqltypes.String):
        # Includes VARCHAR(MAX) in SQL -> VARCHAR() in SQLAlchemy
        if (coltype.length is None and to_mysql) or expand_for_scrubbing:
            return sqltypes.Text()
        if strip_collation:
            return remove_collation(coltype)
        return coltype

    # -------------------------------------------------------------------------
    # Binary
    # -------------------------------------------------------------------------

    # -------------------------------------------------------------------------
    # BIT
    # -------------------------------------------------------------------------
    if typeclass == mssql.base.BIT and to_mysql:
        # MySQL BIT objects have a length attribute.
        return mysql.base.BIT()

    # -------------------------------------------------------------------------
    # TIMESTAMP
    # -------------------------------------------------------------------------
    if (isinstance(coltype, sqltypes.TIMESTAMP) and to_mssql and
            convert_mssql_timestamp):
        # You cannot write explicitly to a TIMESTAMP field in SQL Server; it's
        # used for autogenerated values only.
        # - http://stackoverflow.com/questions/10262426/sql-server-cannot-insert-an-explicit-value-into-a-timestamp-column  # noqa
        # - https://social.msdn.microsoft.com/Forums/sqlserver/en-US/5167204b-ef32-4662-8e01-00c9f0f362c2/how-to-tranfer-a-column-with-timestamp-datatype?forum=transactsql  # noqa
        #   ... suggesting BINARY(8) to store the value.
        # MySQL is more helpful:
        # - http://stackoverflow.com/questions/409286/should-i-use-field-datetime-or-timestamp  # noqa
        return mssql.base.BINARY(8)

    # -------------------------------------------------------------------------
    # Some other type
    # -------------------------------------------------------------------------
    return coltype


# =============================================================================
# Questions about SQLAlchemy column types
# =============================================================================

def is_sqlatype_binary(coltype: TypeEngine) -> bool:
    # Several binary types inherit internally from _Binary, making that the
    # easiest to check.

    # noinspection PyProtectedMember
    return isinstance(coltype, sqltypes._Binary)


def is_sqlatype_date(coltype: TypeEngine) -> bool:
    # isinstance also cheerfully handles multiple inheritance, i.e. if you have
    # class A(object), class B(object), and class C(A, B), followed by x = C(),
    # then all of isinstance(x, A), isinstance(x, B), isinstance(x, C) are True

    # noinspection PyProtectedMember
    return isinstance(coltype, sqltypes._DateAffinity)


def is_sqlatype_integer(coltype: TypeEngine) -> bool:
    return isinstance(coltype, sqltypes.Integer)


def is_sqlatype_numeric(coltype: TypeEngine) -> bool:
    return isinstance(coltype, sqltypes.Numeric)  # includes Float, Decimal


def is_sqlatype_text_of_length_at_least(coltype: TypeEngine,
                                        min_length: int = 1000) -> bool:
    if not isinstance(coltype, sqltypes.String):
        return False  # not a string/text type at all
    if coltype.length is None:
        return True  # string of unlimited length
    return coltype.length >= min_length


def is_sqlatype_text_over_one_char(coltype: TypeEngine) -> bool:
    return is_sqlatype_text_of_length_at_least(coltype, 2)


def does_sqlatype_merit_fulltext_index(coltype: TypeEngine,
                                       min_length: int = 1000) -> bool:
    return is_sqlatype_text_of_length_at_least(coltype, min_length)


def does_sqlatype_require_index_len(coltype: TypeEngine) -> bool:
    # MySQL, at least, requires index length to be specified for BLOB and TEXT
    # columns: http://dev.mysql.com/doc/refman/5.7/en/create-index.html
    if isinstance(coltype, sqltypes.Text):
        return True
    if isinstance(coltype, sqltypes.LargeBinary):
        return True
    return False


# =============================================================================
# Hack in new type
# =============================================================================

def hack_in_mssql_xml_type():
    """
    SQLAlchemy does not support the XML type in SQL Server (mssql).
    Upon reflection, we get:
       sqlalchemy\dialects\mssql\base.py:1921: SAWarning: Did not recognize
       type 'xml' of column '...'

    We will convert anything of type XML into type TEXT.

    """
    log.debug("Adding type 'xml' to SQLAlchemy reflection for dialect 'mssql'")
    mssql.base.ischema_names['xml'] = mssql.base.TEXT
    # http://stackoverflow.com/questions/32917867/sqlalchemy-making-schema-reflection-find-use-a-custom-type-for-all-instances  # noqa

    # print(repr(mssql.base.ischema_names.keys()))
    # print(repr(mssql.base.ischema_names))


# =============================================================================
# Check column definition equality
# =============================================================================

def column_types_equal(a_coltype: TypeEngine, b_coltype: TypeEngine) -> bool:
    # http://stackoverflow.com/questions/34787794/sqlalchemy-column-type-comparison  # noqa
    # IMPERFECT:
    return str(a_coltype) == str(b_coltype)


def columns_equal(a: Column, b: Column) -> bool:
    return (
        a.name == b.name and
        column_types_equal(a.type, b.type) and
        a.nullable == b.nullable
    )


def column_lists_equal(a: List[Column], b: List[Column]) -> bool:
    n = len(a)
    if len(b) != n:
        return False
    for i in range(n):
        if not columns_equal(a[i], b[i]):
            log.debug("Mismatch: {} != {}".format(repr(a[i]), repr(b[i])))
            return False
    return True


def indexes_equal(a: Index, b: Index) -> bool:
    # Unsure.
    return str(a) == str(b)


def index_lists_equal(a: List[Index], b: List[Index]) -> bool:
    n = len(a)
    if len(b) != n:
        return False
    for i in range(n):
        if not indexes_equal(a[i], b[i]):
            log.debug("Mismatch: {} != {}".format(repr(a[i]), repr(b[i])))
            return False
    return True


# =============================================================================
# Tests
# =============================================================================

def test_assert(x, y) -> None:
    try:
        assert x == y
    except AssertionError:
        print("{} should have been {}".format(repr(x), repr(y)))
        raise


def unit_tests() -> None:
    from sqlalchemy.dialects.mssql.base import MSDialect
    from sqlalchemy.dialects.mysql.base import MySQLDialect
    d_mssql = MSDialect()
    d_mysql = MySQLDialect()
    col1 = Column('hello', BigInteger, nullable=True)
    col2 = Column('world', BigInteger,
                  autoincrement=True)  # does NOT generate IDENTITY
    col3 = make_bigint_autoincrement_column('you', d_mssql)
    metadata = MetaData()
    t = Table('mytable', metadata)
    t.append_column(col1)
    t.append_column(col2)
    t.append_column(col3)

    print("Checking Column -> DDL: SQL Server (mssql)")
    test_assert(column_creation_ddl(col1, d_mssql), "hello BIGINT NULL")
    test_assert(column_creation_ddl(col2, d_mssql), "world BIGINT NULL")
    test_assert(column_creation_ddl(col3, d_mssql),
                "you BIGINT NOT NULL IDENTITY(1,1)")

    print("Checking Column -> DDL: MySQL (mysql)")
    test_assert(column_creation_ddl(col1, d_mysql), "hello BIGINT")
    test_assert(column_creation_ddl(col2, d_mysql), "world BIGINT")
    # not col3; unsupported

    print("Checking SQL type -> SQL Alchemy type")
    to_check = [
        # mssql
        ("BIGINT", d_mssql),
        ("NVARCHAR(32)", d_mssql),
        ("NVARCHAR(MAX)", d_mssql),
        ('NVARCHAR(160) COLLATE "Latin1_General_CI_AS"', d_mssql),
        # mysql
        ("BIGINT", d_mssql),
        ("LONGTEXT", d_mysql),
    ]
    for coltype, dialect in to_check:
        print("... {} -> dialect {} -> {}".format(
            repr(coltype),
            repr(dialect.name),
            repr(get_sqla_coltype_from_dialect_str(coltype, dialect))))


if __name__ == '__main__':
    unit_tests()



#!/usr/bin/env python
# crate_anon/crateweb/consent/celery.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import os

from celery import Celery, current_task

# set the default Django settings module for the 'celery' program.
os.environ.setdefault('DJANGO_SETTINGS_MODULE',
                      'crate_anon.crateweb.config.settings')
# RNC: note that the module path must be accessible from the PYTHONPATH,

# from django.conf import settings

app = Celery('consent')

# Using a string here means the worker will not have to
# pickle the object when using Windows.
app.config_from_object('django.conf:settings')

# app.autodiscover_tasks(lambda: settings.INSTALLED_APPS)
# ... looks for "tasks.py" in all apps
# ... REQUIRES that all apps have an __init__.py
#     https://github.com/celery/celery/issues/2523

app.autodiscover_tasks(['crate_anon.crateweb.consent'])  # simpler!


@app.task(bind=True)
def debug_task(self) -> None:
    print('Request: {0!r}'.format(self.request))
    print('Backend: {}'.format(current_task.backend))

"""
===============================================================================
To test:
===============================================================================

1. Launch the Celery worker:
    # no directory change required for an installed package
    # in early testing:
    #   $ cd [...]/crateweb
    #   $ celery --app consent worker --loglevel=debug
    # but with a proper package:
    
    $ celery worker --app crate_anon.crateweb.consent --loglevel debug

2. Call the function asynchronously
    $ manage.py shell
    >>> from consent.celery import debug_task
    >>> debug_task.delay()

    or simply:
    $ python
    >>> from crate_anon.crateweb.consent.celery import debug_task
    >>> debug_task.delay()

===============================================================================
Command-line testing under Windows (2016-05-12):
===============================================================================
(1)
Run crate_launch_celery (which runs the "celery worker" command).
With that running...

(2)
$ python
>>> from crate_anon.crateweb.consent.tasks import add
>>> add(3, 4)  # call immediately
7
>>> result = add.delay(3, 5)  # call via Celery

# AT THAT MOMENT, THE CELERY WORKER PROCESS SHOULD SAY:
[... INFO/MainProcess] Received task: celery.local.add[...]
[... INFO/MainProcess] Task celery.local.add[...] succeeded in ...s: 8

>>> result
<AsyncResult: 48ad8faa-ce75-4a19-bb0e-77feb9567b06>
>>> result.ready()
>>> result.state

(3) When it doesn't work...

$ sudo rabbitmqctl list_queues name messages consumers
    
>>> from crate_anon.crateweb.consent.celery import app
>>> app.control.purge()
# Should report the number of tasks in the queue, after which the rabbitmqctl
# should show an empty queue.

# Using the lengthy "--include" argument to "celery worker", which I was using
# (in launch_celery.py) before creating a proper pip-installed package, seems
# to be unnecessary.

# With "--loglevel info", the task list looks like:
[tasks]
  . crate_anon.crateweb.consent.celery.debug_task
  . crate_anon.crateweb.consent.tasks.add
  . crate_anon.crateweb.consent.tasks.email_rdbm_task
  . crate_anon.crateweb.consent.tasks.finalize_clinician_response
  . crate_anon.crateweb.consent.tasks.process_consent_change
  . crate_anon.crateweb.consent.tasks.process_contact_request
  . crate_anon.crateweb.consent.tasks.process_patient_response
  . crate_anon.crateweb.consent.tasks.resend_email
  . crate_anon.crateweb.consent.tasks.test_email_rdbm_task

With "--loglevel debug", it looks like this:
[tasks]
  . celery.backend_cleanup
  . celery.chain
  . celery.chord
  . celery.chord_unlock
  . celery.chunks
  . celery.group
  . celery.local.add
  . celery.local.email_rdbm_task
  . celery.local.finalize_clinician_response
  . celery.local.process_consent_change
  . celery.local.process_contact_request
  . celery.local.process_patient_response
  . celery.local.resend_email
  . celery.local.test_email_rdbm_task
  . celery.map
  . celery.starmap
  . crate_anon.crateweb.consent.celery.debug_task
  . crate_anon.crateweb.consent.tasks.add
  . crate_anon.crateweb.consent.tasks.email_rdbm_task
  . crate_anon.crateweb.consent.tasks.finalize_clinician_response
  . crate_anon.crateweb.consent.tasks.process_consent_change
  . crate_anon.crateweb.consent.tasks.process_contact_request
  . crate_anon.crateweb.consent.tasks.process_patient_response
  . crate_anon.crateweb.consent.tasks.resend_email
  . crate_anon.crateweb.consent.tasks.test_email_rdbm_task

... but that doesn't stop it working under Linux.


Inspecting workers:
    http://docs.celeryproject.org/en/latest/userguide/workers.html#inspecting-workers  # noqa

Inspecting lots of things:
    pip install flower
    celery -A crate_anon.crateweb.consent flower
    http://localhost:5555/

Celery bug under Windows? https://github.com/celery/celery/issues/897

Upgrade from Celery 3.1.19 to 3.1.23:
    OK! Now we're talking.
    - Flower can now connect to the Celery worker (which reports things in its
      log as Flower does so).
    - Tasks now show up in Flower -> Tasks, marked "Received", with appropriate
      arguments, and with a worker assigned, but the "Started" column remains
      blank.

But then it stopped working again.
- https://github.com/mher/flower/issues/452
  ... indicates that "celery inspect ping" should be doing something, but I'm
  getting "No workers replied within time constraint." On Linux, we get:
    -> celery@wombat: OK
            pong

Improvement if you run the Celery worker with "--concurrency=4" (up from its
default of 1). Flower talks to the nodes. The tasks appear in the worker's
task list. Both "celery status" and "celery inspect ping" work. But tasks
still go into the "Reserved" list.

The "-Ofair" option didn't fix it:
    http://stackoverflow.com/questions/24519559

And then it screwed up again, and became unresponsive to the status things.
Argh!

- Close all Celery things.
- Hard reset of RabbitMQ:
    rabbitmqctl stop_app
    rabbitmqctl reset
    rabbitmqctl start_app
- Start Celery worker(s).
- celery -A crate_anon.crateweb.consent status
    ... now works
- But tasks are still going into the reserved queue.

People have achieved this with Win10:
    https://github.com/hotdogee/django-blast/Windows-Development-Environment-Setup  # noqa

Aha! Using "--pool=solo" made it work!
    https://stackoverflow.com/questions/20309954


===============================================================================
See also tasks.py, with more notes
===============================================================================
"""



#!/usr/bin/env python
# crate_anon/crateweb/userprofile/forms.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

from django.forms import ModelForm
from crate_anon.crateweb.userprofile.models import UserProfile


class UserProfileForm(ModelForm):
    class Meta:
        model = UserProfile
        fields = ['per_page', 'patients_per_page',
                  'line_length', 'collapse_at_len', 'collapse_at_n_lines']



#!/usr/bin/env python
# crate_anon/anonymise/subproc.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import atexit
import logging
from multiprocessing.dummy import Pool  # thread pool
from subprocess import (
    check_call,
    # PIPE,
    Popen,
    # STDOUT,
    TimeoutExpired,
)
import sys
from typing import Any, List

log = logging.getLogger(__name__)

# =============================================================================
# Processes that we're running
# =============================================================================

processes = []  # type: List[Popen]
proc_args_list = []  # type: List[List[str]]
# ... to report back which process failed, if any did


# =============================================================================
# Exiting
# =============================================================================

@atexit.register
def kill_child_processes() -> None:
    timeout_sec = 5
    for p in processes:
        try:
            p.wait(timeout_sec)
        except TimeoutExpired:
            # failed to close
            p.kill()  # you're dead


def fail() -> None:
    print("\nPROCESS FAILED; EXITING ALL\n")
    sys.exit(1)  # will call the atexit handler and kill everything else


# =============================================================================
# Subprocess handling
# =============================================================================

def check_call_process(args: List[str]) -> None:
    log.debug(args)
    check_call(args)


def start_process(args: List[str],
                  stdin: Any = None,
                  stdout: Any = None,
                  stderr: Any = None) -> Popen:
    """

    Args:
        args: program and its arguments, as a list
        stdin: typically None
        stdout: use None to perform no routing, which preserves console colour!
            Otherwise, specify somewhere to route stdout. See subprocess
            documentation. If either is PIPE, you'll need to deal with the
            output.
        stderr: As above. You can use stderr=STDOUT to route stderr to the same
            place as stdout.

    Returns:
        The process object (which is also stored in processes).
    """
    log.debug(args)
    global processes
    global proc_args_list
    proc = Popen(args, stdin=stdin, stdout=stdout, stderr=stderr)
    # proc = Popen(args, stdin=None, stdout=PIPE, stderr=STDOUT)
    # proc = Popen(args, stdin=None, stdout=PIPE, stderr=PIPE)
    # Can't preserve colour: http://stackoverflow.com/questions/13299550/preserve-colored-output-from-python-os-popen  # noqa
    processes.append(proc)
    proc_args_list.append(args)
    return proc


def wait_for_processes(die_on_failure: bool = True,
                       timeout_sec: float = 1) -> None:
    """
    If die_on_failure is True, then whenever a subprocess returns failure,
    all are killed.

    If timeout_sec is None, the function waits for its first process to
    complete, then waits for the second, etc. So a subprocess dying does not
    trigger a full quit instantly (or potentially for ages).

    If timeout_sec is something else, each process is tried for that time;
    if it quits within that time, well and good (successful quit -> continue
    waiting for the others; failure -> kill everything, if die_on_failure);
    if it doesn't, we try the next. That is much more responsive.
    """
    global processes
    global proc_args_list
    n = len(processes)
    Pool(n).map(print_lines, processes)  # in case of PIPE
    something_running = True
    while something_running:
        something_running = False
        for i, p in enumerate(processes):
            try:
                retcode = p.wait(timeout=timeout_sec)
                if retcode == 0:
                    log.info("Process #{} (of {}) exited cleanly".format(i, n))
                if retcode != 0:
                    log.critical(
                        "Process #{} (of {}) exited with return code {} "
                        "(indicating failure); its args were: {}".format(
                            i, n, retcode, repr(proc_args_list[i])))
                    if die_on_failure:
                        log.critical("Exiting top-level process (will kill "
                                     "all other children)")
                        fail()  # exit this process, therefore kill its children  # noqa
            except TimeoutExpired:
                something_running = True
    processes.clear()
    proc_args_list.clear()


def print_lines(process: Popen) -> None:
    out, err = process.communicate()
    if out:
        for line in out.decode("utf-8").splitlines():
            print(line)
    if err:
        for line in err.decode("utf-8").splitlines():
            print(line)


def run_multiple_processes(args_list: List[List[str]],
                           die_on_failure: bool = True) -> None:
    for procargs in args_list:
        start_process(procargs)
    # Wait for them all to finish
    wait_for_processes(die_on_failure=die_on_failure)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-02-28 10:52
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('consent', '0006_auto_20170206_1617'),
    ]

    operations = [
        migrations.AddField(
            model_name='consentmode',
            name='source',
            field=models.CharField(default='crate_user_entry', max_length=20, verbose_name='Source database used for lookup, or crate_user_entry/crate_auto_created'),  # noqa
        ),
        migrations.AlterField(
            model_name='patientlookup',
            name='source_db',
            field=models.CharField(choices=[('dummy_clinical', 'Dummy clinical database for testing'), ('cpft_crs', 'CPFT Care Records System (CRS) 2005-2012'), ('cpft_rio_rcep', 'CPFT RiO 2013- (preprocessed by Servelec RCEP tool)'), ('cpft_rio_crate', 'CPFT RiO 2013- (raw)')], max_length=20, verbose_name='Source database used for lookup'),  # noqa
        ),
    ]



#!/usr/bin/env python
# crate_anon/crateweb/core/views.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import logging
from django.conf import settings
from django.http import HttpResponse
from django.http.request import HttpRequest
from django.shortcuts import render
from crate_anon.crateweb.core.utils import is_developer
from crate_anon.crateweb.research.views import query_context

log = logging.getLogger(__name__)


# =============================================================================
# Home
# =============================================================================

def home(request: HttpRequest) -> HttpResponse:
    # leaflets = [{'key': x[0], 'name': x[1]} for x in Leaflet.LEAFLET_CHOICES]
    # assert False
    # log.critical("IP Address for debug-toolbar: " + request.META['REMOTE_ADDR'])  # noqa
    # log.critical("MIDDLEWARE_CLASSES: {}".format(repr(MIDDLEWARE_CLASSES)))
    context = {
        'nav_on_main_menu': True,
        'is_developer': is_developer(request.user),
        'safety_catch_on': settings.SAFETY_CATCH_ON,
        'developer_email': settings.DEVELOPER_EMAIL,
        # 'leaflets': leaflets,
    }
    context.update(query_context(request))
    return render(request, 'home.html', context)


# =============================================================================
# About
# =============================================================================

def about(request: HttpRequest) -> HttpResponse:
    return render(request, 'about.html')



#!/usr/bin/env python
# crate_anon/crateweb/extra/nhs.py

"""
===============================================================================
    Copyright (C) 2015-2017 Rudolf Cardinal (rudolf@pobox.com).

    This file is part of CRATE.

    CRATE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    CRATE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CRATE. If not, see <http://www.gnu.org/licenses/>.
===============================================================================
"""

import logging
import random
from typing import List, Optional, Union

log = logging.getLogger(__name__)


# =============================================================================
# NHS number validation
# =============================================================================

NHS_DIGIT_WEIGHTINGS = [10, 9, 8, 7, 6, 5, 4, 3, 2]


def nhs_check_digit(ninedigits: Union[str, List[Union[str, int]]]) -> int:
    """
    Calculates an NHS number check digit.
    ninedigits: string or list

    1. Multiply each of the first nine digits by the corresponding
       digit weighting (see NHS_DIGIT_WEIGHTINGS).
    2. Sum the results.
    3. Take remainder after division by 11.
    4. Subtract the remainder from 11
    5. If this is 11, use 0 instead
    If it's 10, the number is invalid
    If it doesn't match the actual check digit, the number is invalid
    """
    if len(ninedigits) != 9 or not all(str(x).isdigit() for x in ninedigits):
        raise ValueError("bad string to nhs_check_digit")
    check_digit = 11 - (sum([
        int(d) * f
        for (d, f) in zip(ninedigits, NHS_DIGIT_WEIGHTINGS)
    ]) % 11)
    # ... % 11 yields something in the range 0-10
    # ... 11 - that yields something in the range 1-11
    if check_digit == 11:
        check_digit = 0
    return check_digit


def is_valid_nhs_number(n: int) -> bool:
    """
    Validates an integer as an NHS number.
    Checksum details are at
        http://www.datadictionary.nhs.uk/version2/data_dictionary/data_field_notes/n/nhs_number_de.asp  # noqa
    """
    if not isinstance(n, int):
        log.debug("is_valid_nhs_number: parameter was not of integer type")
        return False

    s = str(n)
    # Not 10 digits long?
    if len(s) != 10:
        log.debug("is_valid_nhs_number: not 10 digits")
        return False

    main_digits = [int(s[i]) for i in range(9)]
    actual_check_digit = int(s[9])  # tenth digit
    expected_check_digit = nhs_check_digit(main_digits)
    if expected_check_digit == 10:
        log.debug("is_valid_nhs_number: calculated check digit invalid")
        return False
    if expected_check_digit != actual_check_digit:
        log.debug("is_valid_nhs_number: check digit mismatch")
        return False
    # Hooray!
    return True


def generate_random_nhs_number() -> int:
    """Returns a random valid NHS number, as an int."""
    check_digit = 10  # NHS numbers with this check digit are all invalid
    while check_digit == 10:
        digits = [random.randint(1, 9)]  # don't start with a zero
        digits.extend([random.randint(0, 9) for _ in range(8)])
        # ... length now 9
        check_digit = nhs_check_digit(digits)
    # noinspection PyUnboundLocalVariable
    digits.append(check_digit)
    return int("".join([str(d) for d in digits]))


def generate_nhs_number_from_first_9_digits(first9digits: str) -> Optional[int]:
    """
    Returns a valid NHS number, as an int, given the first 9 digits.
    The particular purpose is to make NHS numbers that *look* fake.
    Specifically:
        123456789_ : no; checksum 10
        987654321_ : yes, for 9876543210
        999999999_ : yes, for 9999999999
    """
    if len(first9digits) != 9:
        log.warning("Not 9 digits")
        return None
    try:
        first9int = int(first9digits)
    except (TypeError, ValueError):
        log.warning("Not an integer")
        return None  # not an int
    if len(str(first9int)) != len(first9digits):
        # e.g. leading zeros, or some such
        log.warning("Leading zeros?")
        return None
    check_digit = nhs_check_digit(first9digits)
    if check_digit == 10:  # NHS numbers with this check digit are all invalid
        log.warning("Can't have check digit of 10")
        return None
    return int(first9digits + str(check_digit))
